import pytest

from dbt.tests.adapter.store_test_failures_tests import basic
from dbt.tests.adapter.store_test_failures_tests.test_store_test_failures import (
    StoreTestFailuresBase,
)


TEST_AUDIT_SCHEMA_SUFFIX = "dbt_test__aud"


class TestBigQueryStoreTestFailures(StoreTestFailuresBase):
    @pytest.fixture(scope="function", autouse=True)
    def teardown_method(self, project):
        yield
        relation = project.adapter.Relation.create(
            database=project.database, schema=f"{project.test_schema}_{TEST_AUDIT_SCHEMA_SUFFIX}"
        )

        project.adapter.drop_schema(relation)

    def test_store_and_assert(self, project):
        self.run_tests_store_one_failure(project)
        self.run_tests_store_failures_and_assert(project)


class TestStoreTestFailuresAsInteractions(basic.StoreTestFailuresAsInteractions):
    pass


class TestStoreTestFailuresAsProjectLevelOff(basic.StoreTestFailuresAsProjectLevelOff):
    pass


class TestStoreTestFailuresAsProjectLevelView(basic.StoreTestFailuresAsProjectLevelView):
    pass


class TestStoreTestFailuresAsGeneric(basic.StoreTestFailuresAsGeneric):
    pass


class TestStoreTestFailuresAsProjectLevelEphemeral(basic.StoreTestFailuresAsProjectLevelEphemeral):
    pass


class TestStoreTestFailuresAsExceptions(basic.StoreTestFailuresAsExceptions):
    pass



import pytest
from dbt.tests.util import (
    check_relations_equal,
    get_relation_columns,
    run_dbt,
)
from dbt.tests.adapter.simple_seed.test_seed import SeedConfigBase
from tests.functional.adapter.incremental.seeds import (
    seed_data_csv,
    seed_incremental_overwrite_date_expected_csv,
    seed_incremental_overwrite_day_expected_csv,
    seed_incremental_overwrite_range_expected_csv,
    seed_incremental_overwrite_time_expected_csv,
    seed_merge_expected_csv,
    seed_incremental_overwrite_day_with_time_partition_expected_csv,
)
from tests.functional.adapter.incremental.incremental_strategy_fixtures import (
    merge_range_sql,
    merge_time_sql,
    merge_time_with_require_partition_sql,
    overwrite_date_sql,
    overwrite_day_sql,
    overwrite_day_with_copy_partitions_sql,
    overwrite_partitions_sql,
    overwrite_copy_partitions_with_partitions_sql,
    overwrite_range_sql,
    overwrite_time_sql,
    overwrite_day_with_time_ingestion_sql,
    overwrite_day_with_time_partition_datetime_sql,
    overwrite_static_day_sql,
)


class TestBigQueryScripting(SeedConfigBase):
    @pytest.fixture(scope="class")
    def schema(self):
        return "bigquery_test"

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental_merge_range.sql": merge_range_sql,
            "incremental_merge_time.sql": merge_time_sql,
            "incremental_merge_time_with_require_partition.sql": merge_time_with_require_partition_sql,
            "incremental_overwrite_date.sql": overwrite_date_sql,
            "incremental_overwrite_day.sql": overwrite_day_sql,
            "incremental_overwrite_day_with_copy_partitions.sql": overwrite_day_with_copy_partitions_sql,
            "incremental_overwrite_partitions.sql": overwrite_partitions_sql,
            "incremental_overwrite_copy_partitions_with_partitions.sql": overwrite_copy_partitions_with_partitions_sql,
            "incremental_overwrite_range.sql": overwrite_range_sql,
            "incremental_overwrite_time.sql": overwrite_time_sql,
            "incremental_overwrite_day_with_time_partition.sql": overwrite_day_with_time_ingestion_sql,
            "incremental_overwrite_day_with_time_partition_datetime.sql": overwrite_day_with_time_partition_datetime_sql,
            "incremental_overwrite_static_day.sql": overwrite_static_day_sql,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "data_seed.csv": seed_data_csv,
            "merge_expected.csv": seed_merge_expected_csv,
            "incremental_overwrite_time_expected.csv": seed_incremental_overwrite_time_expected_csv,
            "incremental_overwrite_date_expected.csv": seed_incremental_overwrite_date_expected_csv,
            "incremental_overwrite_day_expected.csv": seed_incremental_overwrite_day_expected_csv,
            "incremental_overwrite_range_expected.csv": seed_incremental_overwrite_range_expected_csv,
            "incremental_overwrite_day_with_time_partition_expected.csv": seed_incremental_overwrite_day_with_time_partition_expected_csv,
        }

    def test__bigquery_assert_incremental_configurations_apply_the_right_strategy(self, project):
        run_dbt(["seed"])
        results = run_dbt()
        assert len(results) == 13

        results = run_dbt()
        assert len(results) == 13

        incremental_strategies = [
            ("incremental_merge_range", "merge_expected"),
            ("incremental_merge_time", "merge_expected"),
            ("incremental_merge_time_with_require_partition_view", "merge_expected"),
            ("incremental_overwrite_time", "incremental_overwrite_time_expected"),
            ("incremental_overwrite_date", "incremental_overwrite_date_expected"),
            ("incremental_overwrite_partitions", "incremental_overwrite_date_expected"),
            (
                "incremental_overwrite_copy_partitions_with_partitions",
                "incremental_overwrite_date_expected",
            ),
            ("incremental_overwrite_day", "incremental_overwrite_day_expected"),
            ("incremental_overwrite_range", "incremental_overwrite_range_expected"),
            (
                "incremental_overwrite_day_with_time_partition_datetime",
                "incremental_overwrite_day_with_time_partition_expected",
            ),
            ("incremental_overwrite_static_day", "incremental_overwrite_day_expected"),
        ]
        db_with_schema = f"{project.database}.{project.test_schema}"
        for incremental_strategy in incremental_strategies:
            created_table = f"{db_with_schema}.{incremental_strategy[0]}"
            expected_table = f"{db_with_schema}.{incremental_strategy[1]}"
            check_relations_equal(project.adapter, [created_table, expected_table])

        # since this table requires a partition filter which check_relations_equal doesn't support extra where clauses
        # we just check column types
        created = get_relation_columns(
            project.adapter, "incremental_overwrite_day_with_copy_partitions"
        )
        expected = get_relation_columns(project.adapter, "incremental_overwrite_day_expected")
        assert created == expected

        created = get_relation_columns(
            project.adapter, "incremental_overwrite_day_with_time_partition"
        )
        expected = get_relation_columns(
            project.adapter, "incremental_overwrite_day_with_time_partition_expected"
        )
        assert created == expected



import pytest
from dbt.tests.util import run_dbt

# This is to test a edge case found in https://github.com/dbt-labs/dbt-bigquery/pull/165/files

tests__get_cols_in_sql = """
{% test get_cols_in(model) %}

  {# The step which causes the issue #}
  {%- set relation = api.Relation.create(identifier=model.table) if execute -%}

  {% set columns = adapter.get_columns_in_relation(relation) %}

  select
    {% for col in columns %}
      {{ col.name }} {{ "," if not loop.last }}
    {% endfor %}

  from {{ model }}
  limit 0

{% endtest %}
"""

models__my_model = """select 1 as id, 'text' as another_col
"""

properties__model_yml = """
version: 2
models:
  - name: my_model
    tests:
      - get_cols_in
"""


class TestIncompleteRelationSetup:
    @pytest.fixture(scope="class")
    def properties(self):
        return {"properties__model_yml.yml": properties__model_yml}

    @pytest.fixture(scope="class")
    def macros(self):
        return {"get_col_in.sql": tests__get_cols_in_sql}

    @pytest.fixture(scope="class")
    def models(self):
        return {"my_model.sql": models__my_model}


class TestIncompleteRelation(TestIncompleteRelationSetup):
    def test_incomplete_relation(self, project):
        run_dbt(["build"])



from dbt.adapters.bigquery.relation import BigQueryRelation
from dbt.adapters.contracts.relation import RelationType


def test_renameable_relation():
    relation = BigQueryRelation.create(
        database="my_db",
        schema="my_schema",
        identifier="my_table",
        type=RelationType.Table,
    )
    assert relation.renameable_relations == frozenset(
        {
            RelationType.Table,
        }
    )



from dbt.tests.adapter.hooks import test_model_hooks as core_base
import pytest


class TestBigQueryPrePostModelHooks(core_base.TestPrePostModelHooks):
    def check_hooks(self, state, project, host, count=1):
        self.get_ctx_vars(state, count=count, project=project)


class TestBigQueryPrePostModelHooksUnderscores(core_base.TestPrePostModelHooksUnderscores):
    def check_hooks(self, state, project, host, count=1):
        self.get_ctx_vars(state, count=count, project=project)


class TestBigQueryHookRefs(core_base.TestHookRefs):
    def check_hooks(self, state, project, host, count=1):
        self.get_ctx_vars(state, count=count, project=project)


class TestBigQueryPrePostModelHooksOnSeeds(core_base.TestPrePostModelHooksOnSeeds):
    def check_hooks(self, state, project, host, count=1):
        self.get_ctx_vars(state, count=count, project=project)

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "models": {},
            "seeds": {
                "+post-hook": [
                    "alter table {{ this }} add column new_col int",
                    "update {{ this }} set new_col = 1 where 1=1",
                ],
                "quote_columns": True,
            },
        }



from typing import Optional

from google.api_core.client_info import ClientInfo
from google.api_core.client_options import ClientOptions
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import aiplatform_v1
from google.cloud.bigquery import Client as BigQueryClient, DEFAULT_RETRY as BQ_DEFAULT_RETRY
from google.cloud.dataproc_v1 import BatchControllerClient, JobControllerClient
from google.cloud.storage import Client as StorageClient
from google.cloud.storage.retry import DEFAULT_RETRY as GCS_DEFAULT_RETRY
from google.oauth2.credentials import Credentials as GoogleCredentials

from dbt.adapters.events.logging import AdapterLogger

import dbt.adapters.bigquery.__version__ as dbt_version
from dbt.adapters.bigquery.credentials import (
    BigQueryCredentials,
    create_google_credentials,
    set_default_credentials,
)


_logger = AdapterLogger("BigQuery")


def create_bigquery_client(credentials: BigQueryCredentials) -> BigQueryClient:
    try:
        return _create_bigquery_client(credentials)
    except DefaultCredentialsError:
        _logger.info("Please log into GCP to continue")
        set_default_credentials()
        return _create_bigquery_client(credentials)


@GCS_DEFAULT_RETRY
def create_gcs_client(credentials: BigQueryCredentials) -> StorageClient:
    return StorageClient(
        project=credentials.execution_project,
        credentials=create_google_credentials(credentials),
    )


# dataproc does not appear to have a default retry like BQ and GCS
def create_dataproc_job_controller_client(credentials: BigQueryCredentials) -> JobControllerClient:
    return JobControllerClient(
        credentials=create_google_credentials(credentials),
        client_options=ClientOptions(api_endpoint=_dataproc_endpoint(credentials)),
    )


# dataproc does not appear to have a default retry like BQ and GCS
def create_dataproc_batch_controller_client(
    credentials: BigQueryCredentials,
) -> BatchControllerClient:
    return BatchControllerClient(
        credentials=create_google_credentials(credentials),
        client_options=ClientOptions(api_endpoint=_dataproc_endpoint(credentials)),
    )


@BQ_DEFAULT_RETRY
def _create_bigquery_client(credentials: BigQueryCredentials) -> BigQueryClient:
    return BigQueryClient(
        credentials.execution_project,
        create_google_credentials(credentials),
        location=getattr(credentials, "location", None),
        client_info=ClientInfo(user_agent=f"dbt-bigquery-{dbt_version.version}"),
        client_options=ClientOptions(
            quota_project_id=credentials.quota_project, api_endpoint=credentials.api_endpoint
        ),
    )


def _dataproc_endpoint(credentials: BigQueryCredentials) -> str:
    return f"{credentials.compute_region}-dataproc.googleapis.com:443"


def create_notebook_client(
    credentials: GoogleCredentials, region: Optional[str]
) -> aiplatform_v1.NotebookServiceClient:
    api_endpoint = f"{region}-aiplatform.googleapis.com"
    notebook_client = aiplatform_v1.NotebookServiceClient(
        credentials=credentials,
        client_options=ClientOptions(api_endpoint),
    )

    return notebook_client



import pytest
from google.api_core.exceptions import NotFound
from dbt.adapters.bigquery.relation import BigQueryRelation
from dbt.tests.util import run_dbt, get_connection, relation_from_name


_INCREMENTAL_MODEL = """
{{
    config(
        materialized="incremental",
        on_schema_change="sync_all_columns"
    )
}}
    select 20 as id, cast('2020-01-01 01:00:00' as datetime) as date_hour union all
    select 40 as id, cast('2020-01-01 02:00:00' as datetime) as date_hour
"""

_INCREMENTAL_MODEL_YAML = """version: 2
models:
- name: test_drop_relation
  columns:
  - name: id
    type: int64
  - name: date_hour
    type: datetime
"""


class BaseIncrementalModelConfig:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_drop_relation.sql": _INCREMENTAL_MODEL,
            "schema.yml": _INCREMENTAL_MODEL_YAML,
        }


class TestIncrementalModel(BaseIncrementalModelConfig):
    def test_incremental_model_succeeds(self, project):
        """
        Steps:
        1. Create the model
        2. Merge into the model using __dbt_tmp table
        3. Assert raises NotFound exception
        """
        results = run_dbt(["run"])
        assert len(results) == 1
        results = run_dbt(["run"])
        assert len(results) == 1
        relation: BigQueryRelation = relation_from_name(
            project.adapter, "test_drop_relation__dbt_tmp"
        )
        adapter = project.adapter
        with pytest.raises(NotFound):
            with get_connection(project.adapter) as conn:
                conn.handle.get_table(
                    adapter.connections.get_bq_table(
                        relation.database, relation.schema, relation.table
                    )
                )



from dbt.adapters.bigquery.relation_configs._base import BigQueryBaseRelationConfig
from dbt.adapters.bigquery.relation_configs._cluster import (
    BigQueryClusterConfig,
    BigQueryClusterConfigChange,
)
from dbt.adapters.bigquery.relation_configs._materialized_view import (
    BigQueryMaterializedViewConfig,
    BigQueryMaterializedViewConfigChangeset,
)
from dbt.adapters.bigquery.relation_configs._options import (
    BigQueryOptionsConfig,
    BigQueryOptionsConfigChange,
)
from dbt.adapters.bigquery.relation_configs._partition import (
    PartitionConfig,
    BigQueryPartitionConfigChange,
)
from dbt.adapters.bigquery.relation_configs._policies import (
    BigQueryIncludePolicy,
    BigQueryQuotePolicy,
)



import pytest
import random
import time
from dbt.tests.util import run_dbt
from dbt.tests.adapter.simple_seed.test_seed import SeedConfigBase
from tests.functional.adapter.simple_bigquery_view.seeds import (
    seed_data_csv,
    seed_merge_expected_csv,
    seed_incremental_overwrite_time_expected_csv,
    seed_incremental_overwrite_date_expected_csv,
    seed_incremental_overwrite_day_expected_csv,
    seed_incremental_overwrite_range_expected_csv,
)
from tests.functional.adapter.simple_bigquery_view.fixtures import (
    clustered_model_sql,
    funky_case_sql,
    labeled_model_sql,
    multi_clustered_model_sql,
    partitioned_model_sql,
    sql_header_model_sql,
    sql_header_model_incr_sql,
    sql_header_model_incr_insert_overwrite_sql,
    sql_header_model_incr_insert_overwrite_static_sql,
    tabel_model_sql,
    view_model_sql,
    schema_yml,
    test_creation_sql,
    test_int_inference_sql,
    test_project_for_job_id_sql,
    wrapped_macros_sql,
)


class BaseBigQueryRun(SeedConfigBase):
    @pytest.fixture(scope="class")
    def schema(self):
        return "bigquery_test"

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "test_creation.sql": test_creation_sql,
            "test_int_inference.sql": test_int_inference_sql,
            "test_project_for_job_id.sql": test_project_for_job_id_sql,
            "wrapped_macros.sql": wrapped_macros_sql,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "clustered_model.sql": clustered_model_sql,
            "fUnKyCaSe.sql": funky_case_sql,
            "labeled_model.sql": labeled_model_sql,
            "multi_clustered_model.sql": multi_clustered_model_sql,
            "partitioned_model.sql": partitioned_model_sql,
            "sql_header_model.sql": sql_header_model_sql,
            "sql_header_model_incr.sql": sql_header_model_incr_sql,
            "sql_header_model_incr_insert_overwrite.sql": sql_header_model_incr_insert_overwrite_sql,
            "sql_header_model_incr_insert_overwrite_static.sql": sql_header_model_incr_insert_overwrite_static_sql,
            "table_model.sql": tabel_model_sql,
            "view_model.sql": view_model_sql,
            "schema.yml": schema_yml,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "data_seed.csv": seed_data_csv,
            "merge_expected.csv": seed_merge_expected_csv,
            "incremental_overwrite_time_expected.csv": seed_incremental_overwrite_time_expected_csv,
            "incremental_overwrite_date_expected.csv": seed_incremental_overwrite_date_expected_csv,
            "incremental_overwrite_day_expected.csv": seed_incremental_overwrite_day_expected_csv,
            "incremental_overwrite_range_expected.csv": seed_incremental_overwrite_range_expected_csv,
        }

    def assert_nondupes_pass(self, project):
        # The 'dupe' model should fail, but all others should pass
        test_results = run_dbt(["test"], expect_pass=False)

        for test_result in test_results:
            if "dupe" in test_result.node.name:
                assert test_result.status == "fail"
                assert not test_result.skipped
                assert test_result.failures > 0

            # assert that actual tests pass
            else:
                assert test_result.status == "pass"
                assert not test_result.skipped
                assert test_result.failures == 0


class TestSimpleBigQueryRun(BaseBigQueryRun):
    def test__bigquery_simple_run(self, project):
        # make sure seed works twice. Full-refresh is a no-op
        run_dbt(["seed"])
        run_dbt(["seed", "--full-refresh"])

        results = run_dbt()
        # Bump expected number of results when adding new model
        assert len(results) == 11
        self.assert_nondupes_pass(project)


class TestUnderscoreBigQueryRun(BaseBigQueryRun):
    prefix = "_test{}{:04}".format(int(time.time()), random.randint(0, 9999))

    def test_bigquery_run_twice(self, project):
        run_dbt(["seed"])
        results = run_dbt()
        assert len(results) == 11

        results = run_dbt()
        assert len(results) == 11

        self.assert_nondupes_pass(project)



SCHEMA_YML = """version: 2
sources:
  - name: test_source
    freshness:
      warn_after: {count: 10, period: hour}
      error_after: {count: 1, period: day}
    schema: "{{ env_var('DBT_GET_LAST_RELATION_TEST_SCHEMA') }}"
    tables:
      - name: test_source
"""

BATCH_SCHEMA_YML = """
sources:
  - name: test_source
    freshness:
      warn_after: {count: 10, period: hour}
      error_after: {count: 1, period: day}
    schema: "{{ target.schema }}"
    tables:
      - name: test_table
      - name: test_table2
      - name: test_table_with_loaded_at_field
        loaded_at_field: my_loaded_at_field
"""

SEED_TEST_SOURCE_CSV = """
id,name
1,Martin
2,Jeter
3,Ruth
4,Gehrig
5,DiMaggio
6,Torre
7,Mantle
8,Berra
9,Maris
""".strip()



import pytest
from dbt.tests.util import run_dbt, run_dbt_and_capture
import dbt.exceptions

incremental_change_history_enabled_model = """
{{
    config(
        materialized='incremental',
        enable_change_history=True
    )
}}
select 1 as id
"""

incremental_change_history_disabled_model = """
{{
    config(
        materialized='incremental',
        enable_change_history=False
    )
}}
select 1 as id
"""

table_change_history_enabled_model = """
{{
    config(
        materialized='table',
        enable_change_history=True
    )
}}
select 1 as id
"""

view_change_history_enabled_model = """
{{
    config(
        materialized='view',
        enable_change_history=True
    )
}}
select 1 as id
"""


class TestChangeHistorySuccessful:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental_change_history_enabled.sql": incremental_change_history_enabled_model,
            "incremental_change_history_disabled.sql": incremental_change_history_disabled_model,
            "table_change_history_enabled.sql": table_change_history_enabled_model,
        }

    def get_is_change_history_enabled(self, project, table_name):
        sql = f"""
        select is_change_history_enabled
        from `{project.database}.{project.test_schema}.INFORMATION_SCHEMA.TABLES`
        where table_name = '{table_name}'
        """
        results = project.run_sql(sql, fetch="one")
        return results[0] if results else None

    def test_change_history_options(self, project):
        results = run_dbt(["run"])
        assert len(results) == 3

        assert (
            self.get_is_change_history_enabled(project, "incremental_change_history_enabled")
            == "YES"
        )
        assert (
            self.get_is_change_history_enabled(project, "incremental_change_history_disabled")
            == "NO"
        )
        assert self.get_is_change_history_enabled(project, "table_change_history_enabled") == "YES"


class TestChangeHistoryFail:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "view_with_flag.sql": view_change_history_enabled_model,
        }

    def test_view_with_flag_fails(self, project):
        _, stdout = run_dbt_and_capture(["run", "--select", "view_with_flag"], expect_pass=False)
        assert "`enable_change_history` is not supported for views" in stdout



from datetime import datetime
import random
from typing import Optional

from dbt.tests.util import (
    get_connection,
    get_model_file,
    relation_from_name,
    set_model_file,
)


def query_relation_type(project, name: str) -> Optional[str]:
    relation = relation_from_name(project.adapter, name)

    with get_connection(project.adapter):
        results = project.adapter.get_bq_table(relation)

    assert results is not None, f"Relation {relation} not found"
    return results.table_type


def unique_prefix():
    # create a directory name that will be unique per test session
    _randint = random.randint(0, 9999)
    _runtime_timedelta = datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0)
    _runtime = (int(_runtime_timedelta.total_seconds() * 1e6)) + _runtime_timedelta.microseconds
    return f"test{_runtime}{_randint:04}"


def update_model(project, name: str, model: str) -> str:
    relation = relation_from_name(project.adapter, name)
    original_model = get_model_file(project, relation)
    set_model_file(project, relation, model)
    return original_model



import json
import pytest
import re
from dbt.tests.util import run_dbt, run_dbt_and_capture

# This is a short term hack, we need to go back
# and make adapter implementations of:
# https://github.com/dbt-labs/dbt-core/pull/6330


_COMMENT_RE = re.compile(r'/\*\s*{[^}]*"dbt_version"\s*:\s*"[^"]+"[^}]*}\s*\*/')

_INCREMENTAL_MODEL = """
{{
    config(
        materialized="incremental",
    )
}}

{% if not is_incremental() %}

    select
        10 as id, cast('2020-01-01 01:00:00' as datetime) as date_hour
    union all select
        30 as id, cast('2020-01-01 02:00:00' as datetime) as date_hour

{% else %}

    select
        20 as id, cast('2020-01-01 01:00:00' as datetime) as date_hour
    union all select
        40 as id, cast('2020-01-01 02:00:00' as datetime) as date_hour

{% endif %}
-- Test Comment To Prevent Recurrence of
--     https://github.com/dbt-labs/dbt-core/issues/6485
"""

INCREMENTAL_MODEL_COPY_PARTITIONS = """
{{
  config(
      materialized='incremental',
      incremental_strategy='insert_overwrite',
      partition_by={
          'field': '_partition',
          'granularity': 'day',
          'data_type': 'timestamp',
          'time_ingestion_partitioning': True,
          'copy_partitions': True,
      },
      on_schema_change='append_new_columns'
  )
}}
SELECT
  timestamp_trunc(current_timestamp(), day) AS _partition,
  'some value'                               AS col1
"""


class BaseIncrementalModelConfig:
    @pytest.fixture(scope="class")
    def models(self):
        return {"test_incremental.sql": _INCREMENTAL_MODEL}


class TestIncrementalModel(BaseIncrementalModelConfig):
    def test_incremental_model_succeeds(self, project):
        results = run_dbt(["run"])
        assert len(results) == 1
        results = run_dbt(["run"])
        assert len(results) == 1


class TestAllQueriesHaveDbtComment:
    @pytest.fixture(scope="class")
    def models(self):
        return {"my_incremental_model.sql": INCREMENTAL_MODEL_COPY_PARTITIONS}

    def _extract_executed_sql(self, raw_logs: str) -> list[str]:
        """
        Return every SQL script that dbt 1.4+ actually sent to BigQuery.

        In JSON logs each statement is logged by an event whose `data`
        payload is a dict containing a key `"sql"`.
        """
        scripts: list[str] = []
        for line in raw_logs.splitlines():
            try:
                parsed = json.loads(line)
            except json.JSONDecodeError:
                continue

            data = parsed.get("data")
            if isinstance(data, dict) and "sql" in data:
                sql = str(data["sql"]).strip()
                if sql:
                    scripts.append(sql)
        return scripts

    def _has_structured_comment(self, sql: str) -> bool:
        """True iff the first non-blank line is the structured dbt comment."""
        first_line = sql.lstrip().splitlines()[0]
        return bool(_COMMENT_RE.fullmatch(first_line))

    def test_every_query_has_comment(self, project):
        run_dbt(["run"])
        _, raw_logs = run_dbt_and_capture(["--debug", "--log-format=json", "run"])

        executed_sqls = self._extract_executed_sql(raw_logs)
        assert executed_sqls, "No SQL was captured from the dbt logs"

        missing = [sql for sql in executed_sqls if not self._has_structured_comment(sql)]

        assert not missing, (
            f"{len(missing)} queries are missing structured dbt comments.\n\n"
            + "\n\n---\n\n".join(missing)
        )



import pytest

from dbt.tests.util import get_connection
from google.cloud.bigquery import Client, DatasetReference, TableReference
from google.api_core.exceptions import NotFound


@pytest.mark.parametrize("table_name", ["this_table_does_not_exist"])
def test_get_table_does_not_exist(project, table_name):
    """
    TODO: replace dbt project methods with direct connection instantiation
    """
    with get_connection(project.adapter) as conn:
        client: Client = conn.handle
        dataset_ref = DatasetReference(project.database, project.test_schema)
        table_ref = TableReference(dataset_ref, table_name)
        with pytest.raises(NotFound):
            client.get_table(table_ref)



SINGLE_RECORD = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table"
    )

    df = pd.DataFrame(
        [
            {"column_name": {"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}},
        ]
    )

    return df
"""


MULTI_RECORD = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table",
    )

    df = pd.DataFrame(
        [
            {"column_name": [{"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}]},
        ]
    )

    return df
"""


ORC_FORMAT = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table",
        intermediate_format="orc",
    )

    df = pd.DataFrame(
        [
            {"column_name": [{"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}]},
        ]
    )

    return df
"""


ENABLE_LIST_INFERENCE = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table",
        enable_list_inference="true",
    )

    df = pd.DataFrame(
        [
            {"column_name": [{"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}]},
        ]
    )

    return df
"""


ENABLE_LIST_INFERENCE_PARQUET_FORMAT = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table",
        enable_list_inference="true",
        intermediate_format="parquet",
    )

    df = pd.DataFrame(
        [
            {"column_name": [{"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}]},
        ]
    )

    return df
"""


DISABLE_LIST_INFERENCE_ORC_FORMAT = """
import pandas as pd

def model(dbt, session):

    dbt.config(
        submission_method="serverless",
        materialized="table",
        enable_list_inference="false",
        intermediate_format="orc",
    )

    df = pd.DataFrame(
        [
            {"column_name": [{"name": "hello", "my_list": ["h", "e", "l", "l", "o"]}]},
        ]
    )

    return df

"""



import pytest

from dbt.tests.util import run_dbt
from dbt.tests.adapter.materialized_view.basic import MaterializedViewBasic

from tests.functional.adapter.materialized_view_tests._mixin import BigQueryMaterializedViewMixin
from tests.functional.adapter.materialized_view_tests import _files


class TestBigqueryMaterializedViewsBasic(BigQueryMaterializedViewMixin, MaterializedViewBasic):
    def test_view_replaces_materialized_view(self, project, my_materialized_view):
        """
        We don't support replacing a view with another object in dbt-bigquery unless we use --full-refresh
        """
        run_dbt(["run", "--models", my_materialized_view.identifier])
        assert self.query_relation_type(project, my_materialized_view) == "materialized_view"

        self.swap_materialized_view_to_view(project, my_materialized_view)

        # add --full-refresh
        run_dbt(["run", "--models", my_materialized_view.identifier, "--full-refresh"])
        assert self.query_relation_type(project, my_materialized_view) == "view"

    @pytest.mark.skip(
        "It looks like BQ updates the materialized view almost immediately, which fails this test."
    )
    def test_materialized_view_only_updates_after_refresh(
        self, project, my_materialized_view, my_seed
    ):
        pass


class TestMaterializedViewRerun:
    """
    This addresses: https://github.com/dbt-labs/dbt-bigquery/issues/1007

    This effectively tests that defaults get properly set so that the run is idempotent.
    If the defaults are not properly set, changes could appear when there are no changes
    and cause unexpected scenarios.
    """

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        return {"my_minimal_materialized_view.sql": _files.MY_MINIMAL_MATERIALIZED_VIEW}

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": _files.MY_SEED}

    def test_minimal_materialized_view_is_idempotent(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        run_dbt(["run"])



import os
import pytest
from unittest import mock

from dbt.adapters.bigquery.impl import BigQueryAdapter
from dbt.adapters.capability import Capability, CapabilityDict
from dbt.tests.util import run_dbt
from dbt.cli.main import dbtRunner

from tests.functional.adapter.sources_freshness_tests import files


class TestGetLastRelationModified:
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"test_source.csv": files.SEED_TEST_SOURCE_CSV}

    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": files.SCHEMA_YML}

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        # we need the schema name for the sources section
        os.environ["DBT_GET_LAST_RELATION_TEST_SCHEMA"] = project.test_schema
        run_dbt(["seed"])
        yield
        del os.environ["DBT_GET_LAST_RELATION_TEST_SCHEMA"]

    def test_get_last_relation_modified(self, project):
        results = run_dbt(["source", "freshness"])
        assert len(results) == 1
        result = results[0]
        assert result.status == "pass"


class TestGetLastRelationModifiedBatch:
    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": files.BATCH_SCHEMA_YML}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "flags": {
                "bigquery_use_batch_source_freshness": True,
            }
        }

    def get_freshness_result_for_table(self, table_name, results):
        for result in results:
            if result.node.name == table_name:
                return result
        return None

    def test_get_last_relation_modified_batch(self, project, unique_schema, logs_dir):
        project.run_sql(
            f"create table {unique_schema}.test_table as (select 1 as id, 'test' as name);"
        )
        project.run_sql(
            f"create table {unique_schema}.test_table2 as (select 1 as id, 'test' as name);"
        )
        project.run_sql(
            f"create table {unique_schema}.test_table_with_loaded_at_field as (select 1 as id, timestamp '2009-09-15 10:59:43' as my_loaded_at_field);"
        )

        runner = dbtRunner()
        freshness_results_batch = runner.invoke(["source", "freshness"]).result

        assert len(freshness_results_batch) == 3
        test_table_batch_result = self.get_freshness_result_for_table(
            "test_table", freshness_results_batch
        )
        test_table2_batch_result = self.get_freshness_result_for_table(
            "test_table2", freshness_results_batch
        )
        test_table_with_loaded_at_field_batch_result = self.get_freshness_result_for_table(
            "test_table_with_loaded_at_field", freshness_results_batch
        )

        log_file = os.path.join(logs_dir, "dbt.log")
        with open(log_file, "r") as f:
            log = f.read()
            assert "INFORMATION_SCHEMA.TABLE_STORAGE" in log

        # Remove TableLastModifiedMetadataBatch and run freshness on same input without batch strategy
        capabilities_no_batch = CapabilityDict(
            {
                capability: support
                for capability, support in BigQueryAdapter.capabilities().items()
                if capability != Capability.TableLastModifiedMetadataBatch
            }
        )
        with mock.patch.object(
            BigQueryAdapter, "capabilities", return_value=capabilities_no_batch
        ):
            freshness_results = runner.invoke(["source", "freshness"]).result

        assert len(freshness_results) == 3
        test_table_result = self.get_freshness_result_for_table("test_table", freshness_results)
        test_table2_result = self.get_freshness_result_for_table("test_table2", freshness_results)
        test_table_with_loaded_at_field_result = self.get_freshness_result_for_table(
            "test_table_with_loaded_at_field", freshness_results
        )

        # assert results between batch vs non-batch freshness strategy are equivalent
        assert test_table_result.status == test_table_batch_result.status
        assert test_table_result.max_loaded_at == test_table_batch_result.max_loaded_at

        assert test_table2_result.status == test_table2_batch_result.status
        assert test_table2_result.max_loaded_at == test_table2_batch_result.max_loaded_at

        assert (
            test_table_with_loaded_at_field_batch_result.status
            == test_table_with_loaded_at_field_result.status
        )
        assert (
            test_table_with_loaded_at_field_batch_result.max_loaded_at
            == test_table_with_loaded_at_field_result.max_loaded_at
        )


class TestGetLastRelationModifiedBatchLegacy:
    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": files.BATCH_SCHEMA_YML}

    def test_get_last_relation_modified_batch(self, project, unique_schema, logs_dir):
        project.run_sql(
            f"create table {unique_schema}.test_table as (select 1 as id, 'test' as name);"
        )
        project.run_sql(
            f"create table {unique_schema}.test_table2 as (select 1 as id, 'test' as name);"
        )
        project.run_sql(
            f"create table {unique_schema}.test_table_with_loaded_at_field as (select 1 as id, timestamp '2009-09-15 10:59:43' as my_loaded_at_field);"
        )

        runner = dbtRunner()
        runner.invoke(["source", "freshness"])

        log_file = os.path.join(logs_dir, "dbt.log")
        with open(log_file, "r") as f:
            log = f.read()
            assert "INFORMATION_SCHEMA.TABLE_STORAGE" not in log



import pytest
from dbt.tests.adapter.dbt_show.test_dbt_show import (
    BaseShowSqlHeader,
    BaseShowLimit,
    BaseShowDoesNotHandleDoubleLimit,
)

from dbt.tests.util import run_dbt

model_with_json_struct = """
    select *
    from (
        select
  struct<
    k array<
        struct<c1 int64, c2 json>
      >
  >(
    [
      struct(
        1 as c1,
        to_json(struct(1 as a)) as c2
      )
    ]
  )
  as v
    )
    """

model_with_null_json_struct = """
    select (struct<k json>(null)) as null_struct
"""


class TestBigQueryShowLimit(BaseShowLimit):
    pass


class TestBigQueryShowSqlHeader(BaseShowSqlHeader):
    pass


# Added to check if dbt show works with JSON struct
# Addresses: https://github.com/dbt-labs/dbt-bigquery/issues/972
class TestBigQueryShowSqlWorksWithJSONStruct:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "json_struct_model.sql": model_with_json_struct,
            "null_json_struct_model.sql": model_with_null_json_struct,
        }

    def test_sql_header(self, project):
        run_dbt(["show", "--select", "json_struct_model", "-d"])

    def test_show_with_null_json_struct(self, project):
        run_dbt(["show", "--select", "null_json_struct_model"])


class TestBigQueryShowDoesNotHandleDoubleLimit(BaseShowDoesNotHandleDoubleLimit):
    DATABASE_ERROR_MESSAGE = "Syntax error: Expected end of input but got keyword LIMIT"



import pytest

from pathlib import Path

from dbt.tests.util import run_dbt, rm_file, write_file, check_relations_equal

from dbt.tests.adapter.simple_copy.test_simple_copy import SimpleCopyBase

from tests.functional.adapter.simple_copy.fixtures import (
    _MODELS_INCREMENTAL_UPDATE_COLS,
    _SEEDS__SEED_MERGE_COLS_INITIAL,
    _SEEDS__SEED_MERGE_COLS_UPDATE,
    _SEEDS__SEED_MERGE_COLS_EXPECTED_RESULT,
)


class TestSimpleCopyBase(SimpleCopyBase):
    pass


class TestIncrementalMergeColumns:
    @pytest.fixture(scope="class")
    def models(self):
        return {"incremental_update_cols.sql": _MODELS_INCREMENTAL_UPDATE_COLS}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": _SEEDS__SEED_MERGE_COLS_INITIAL}

    def test_incremental_merge_columns(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])

        main_seed_file = project.project_root / Path("seeds") / Path("seed.csv")
        expected_seed_file = project.project_root / Path("seeds") / Path("expected_result.csv")
        rm_file(main_seed_file)
        write_file(_SEEDS__SEED_MERGE_COLS_UPDATE, main_seed_file)
        write_file(_SEEDS__SEED_MERGE_COLS_EXPECTED_RESULT, expected_seed_file)

        run_dbt(["seed"])
        run_dbt(["run"])
        check_relations_equal(project.adapter, ["incremental_update_cols", "expected_result"])



from dbt.tests.util import run_dbt
from dbt.tests.adapter.dbt_debug.test_dbt_debug import BaseDebug, BaseDebugProfileVariable


class TestDebugBigQuery(BaseDebug):
    def test_ok_bigquery(self, project):
        run_dbt(["debug"])
        assert "ERROR" not in self.capsys.readouterr().out


class TestDebugProfileVariableBigQuery(BaseDebugProfileVariable):
    def test_ok_bigquery(self, project):
        run_dbt(["debug"])
        assert "ERROR" not in self.capsys.readouterr().out



import pytest
import yaml
from dbt.tests.util import run_dbt
from dbt.tests.adapter.column_types.test_column_types import BaseColumnTypes
from dbt.tests.adapter.column_types.fixtures import macro_test_is_type_sql
from tests.functional.adapter.column_types.fixtures import (
    _MACRO_TEST_ALTER_COLUMN_TYPE,
    _MODEL_ALT_SQL,
    _ALT_SCHEMA_YML,
)


class BaseAlterColumnTypes(BaseColumnTypes):
    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "test_is_type.sql": macro_test_is_type_sql,
            "test_alter_column_type.sql": _MACRO_TEST_ALTER_COLUMN_TYPE,
        }

    def run_and_alter_and_test(self, alter_column_type_args):
        results = run_dbt(["run"])
        assert len(results) == 1
        run_dbt(["run-operation", "test_alter_column_type", "--args", alter_column_type_args])
        results = run_dbt(["test"])
        assert len(results) == 1


class TestBigQueryAlterColumnTypes(BaseAlterColumnTypes):
    @pytest.fixture(scope="class")
    def models(self):
        return {"model.sql": _MODEL_ALT_SQL, "schema.yml": _ALT_SCHEMA_YML}

    def test_bigquery_alter_column_types(self, project):
        alter_column_type_args = yaml.safe_dump(
            {"model_name": "model", "column_name": "int64_col", "new_column_type": "string"}
        )

        self.run_and_alter_and_test(alter_column_type_args)



MY_SEED = """
id,value,record_valid_date
1,100,2023-01-01 00:00:00
2,200,2023-01-02 00:00:00
3,300,2023-01-02 00:00:00
""".strip()


MY_TABLE = """
{{ config(
    materialized='table',
) }}
select *
from {{ ref('my_seed') }}
"""


MY_VIEW = """
{{ config(
    materialized='view',
) }}
select *
from {{ ref('my_seed') }}
"""


MY_MATERIALIZED_VIEW = """
{{ config(
    materialized='materialized_view',
) }}
select *
from {{ ref('my_table') }}
"""



import pytest
from dbt.tests.util import relation_from_name, get_connection, run_dbt

from dbt.adapters.bigquery import BigQueryRelation

_FIELD_DESCRIPTION_MODEL = """{{
  config(
    materialized='table',
    persist_docs={ 'columns': true }
  )
}}

select
  1 field
"""
_FIELD_DESCRIPTION_MODEL_NAME = "field_description_model"
_FIELD_DESCRIPTION = "this is not a field"
_FIELD_DESCRIPTION_MODEL_YML = """
version: 2

models:
- name: field_description_model
  columns:
  - name: field
    description: '{{ var("field_description") }}'
"""


class TestBigqueryUpdateColumnDescription:
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"config-version": 2, "vars": {"field_description": _FIELD_DESCRIPTION}}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            f"{_FIELD_DESCRIPTION_MODEL_NAME}.sql": _FIELD_DESCRIPTION_MODEL,
            "schema.yml": _FIELD_DESCRIPTION_MODEL_YML,
        }

    def test_bigquery_update_column_description(self, project):
        results = run_dbt(["run"])
        assert len(results) == 1
        relation: BigQueryRelation = relation_from_name(
            project.adapter, _FIELD_DESCRIPTION_MODEL_NAME
        )
        adapter = project.adapter
        with get_connection(project.adapter) as conn:
            table = conn.handle.get_table(
                adapter.connections.get_bq_table(
                    relation.database, relation.schema, relation.table
                )
            )
            for schema_field in table.schema:
                assert schema_field.description == _FIELD_DESCRIPTION



import inspect
import json
import re
import time
from typing import Any, Dict, Optional, Union
import uuid

from dbt.adapters.base import PythonJobHelper
from dbt.adapters.bigquery import BigQueryCredentials
from dbt.adapters.bigquery.clients import (
    create_dataproc_batch_controller_client,
    create_dataproc_job_controller_client,
    create_gcs_client,
    create_notebook_client,
)
from dbt.adapters.bigquery.credentials import (
    BigQueryConnectionMethod,
    create_google_credentials,
    DataprocBatchConfig,
)
from dbt.adapters.bigquery.retry import RetryFactory
from dbt.adapters.events.logging import AdapterLogger
from dbt_common.exceptions import DbtRuntimeError

from google.api_core.operation import Operation
from google.auth.transport.requests import Request
from google.cloud import aiplatform_v1
from google.cloud.aiplatform import gapic as aiplatform_gapic
from google.cloud.dataproc_v1 import CreateBatchRequest, Job, RuntimeConfig
from google.cloud.dataproc_v1.types.batches import Batch
from google.protobuf.json_format import ParseDict
import nbformat

_logger = AdapterLogger("BigQuery")


# Google Cloud usually automatically creates VPC Network & Subnetwork named
# "default" in the project when enabling the Compute Engine API. We will use
# the "default" network to create a default runtime template when needed.
_NETWORK_NAME = "default"
_SUBNETWORK_NAME = "default"
_DEFAULT_JAR_FILE_URI = "gs://spark-lib/bigquery/spark-bigquery-with-dependencies_2.13-0.34.0.jar"
# For BigFrames models, the default timeout is set to 1 hour (60 * 60 seconds).
# This differs from other Python models because the typical 5-minute timeout
# is insufficient for BigFrames processing.
_DEFAULT_BIGFRAMES_TIMEOUT = 60 * 60
# Time interval in seconds between successive polling attempts to check the
# notebook job's status in BigFrames mode.
_COLAB_POLL_INTERVAL = 30
# Suffix used by service accounts.
_SERVICE_ACCOUNT_SUFFIX = "iam.gserviceaccount.com"


class _BigQueryPythonHelper(PythonJobHelper):
    def __init__(self, parsed_model: Dict, credentials: BigQueryCredentials) -> None:
        self._storage_client = create_gcs_client(credentials)
        self._project = credentials.execution_project
        self._region = credentials.compute_region
        # validate all additional stuff for python is set
        for required_config in ["compute_region", "gcs_bucket"]:
            if not getattr(credentials, required_config):
                raise ValueError(
                    f"Need to supply {required_config} in profile to submit python job"
                )

        schema = parsed_model["schema"]
        identifier = parsed_model["alias"]
        self._model_file_name = f"{schema}/{identifier}.py"
        self._gcs_bucket = credentials.gcs_bucket
        self._gcs_path = f"gs://{credentials.gcs_bucket}/{self._model_file_name}"
        self._parsed_model = parsed_model

        # set retry policy, default to timeout after 24 hours
        retry = RetryFactory(credentials)
        self._polling_retry = retry.create_polling(
            model_timeout=parsed_model["config"].get("timeout")
        )

    def _write_to_gcs(self, compiled_code: str) -> None:
        bucket = self._storage_client.get_bucket(self._gcs_bucket)
        blob = bucket.blob(self._model_file_name)
        blob.upload_from_string(compiled_code)

    def _get_batch_id(self) -> str:
        model = self._parsed_model
        default_batch_id = str(uuid.uuid4())
        return model["config"].get("batch_id", default_batch_id)


class ClusterDataprocHelper(_BigQueryPythonHelper):
    def __init__(self, parsed_model: Dict, credentials: BigQueryCredentials) -> None:
        super().__init__(parsed_model, credentials)

        self._job_controller_client = create_dataproc_job_controller_client(credentials)
        self._cluster_name = parsed_model["config"].get(
            "dataproc_cluster_name", credentials.dataproc_cluster_name
        )

        if not self._cluster_name:
            raise ValueError(
                "Need to supply dataproc_cluster_name in profile or config to submit python job with cluster submission method"
            )

    def submit(self, compiled_code: str) -> Job:
        _logger.debug(f"Submitting cluster job to: {self._cluster_name}")

        self._write_to_gcs(compiled_code)

        request = {
            "project_id": self._project,
            "region": self._region,
            "job": {
                "placement": {"cluster_name": self._cluster_name},
                "pyspark_job": {
                    "main_python_file_uri": self._gcs_path,
                },
            },
        }

        # submit the job
        operation = self._job_controller_client.submit_job_as_operation(request)

        # wait for the job to complete
        response: Job = operation.result(polling=self._polling_retry)

        if response.status.state == 6:
            raise ValueError(response.status.details)

        return response


class ServerlessDataProcHelper(_BigQueryPythonHelper):
    def __init__(self, parsed_model: Dict, credentials: BigQueryCredentials) -> None:
        super().__init__(parsed_model, credentials)

        self._batch_controller_client = create_dataproc_batch_controller_client(credentials)
        self._jar_file_uri = parsed_model["config"].get("jar_file_uri", _DEFAULT_JAR_FILE_URI)
        self._dataproc_batch = credentials.dataproc_batch

    def submit(self, compiled_code: str) -> Batch:
        _logger.debug(f"Submitting batch job with id: {self._get_batch_id()}")

        self._write_to_gcs(compiled_code)

        request = CreateBatchRequest(
            parent=f"projects/{self._project}/locations/{self._region}",
            batch=self._create_batch(),
            batch_id=self._get_batch_id(),
        )

        # submit the batch
        operation = self._batch_controller_client.create_batch(request)

        # wait for the batch to complete
        response: Batch = operation.result(polling=self._polling_retry)

        return response

    def _create_batch(self) -> Batch:
        # create the Dataproc Serverless job config
        # need to pin dataproc version to 1.1 as it now defaults to 2.0
        # https://cloud.google.com/dataproc-serverless/docs/concepts/properties
        # https://cloud.google.com/dataproc-serverless/docs/reference/rest/v1/projects.locations.batches#runtimeconfig
        batch = Batch(
            {
                "runtime_config": RuntimeConfig(
                    version="1.1",
                    properties={
                        "spark.executor.instances": "2",
                    },
                ),
                "pyspark_batch": {
                    "main_python_file_uri": self._gcs_path,
                    "jar_file_uris": [self._jar_file_uri],
                },
            }
        )

        # Apply configuration from dataproc_batch key, possibly overriding defaults.
        if self._dataproc_batch:
            batch = _update_batch_from_config(self._dataproc_batch, batch)

        return batch


def _update_batch_from_config(
    config_dict: Union[Dict, DataprocBatchConfig], target: Batch
) -> Batch:
    try:
        # updates in place
        ParseDict(config_dict, target._pb)
    except Exception as e:
        docurl = (
            "https://cloud.google.com/dataproc-serverless/docs/reference/rpc/google.cloud.dataproc.v1"
            "#google.cloud.dataproc.v1.Batch"
        )
        raise ValueError(
            f"Unable to parse dataproc_batch as valid batch specification. See {docurl}. {str(e)}"
        ) from e
    return target


class BigFramesHelper(_BigQueryPythonHelper):

    def __init__(self, parsed_model: Dict, credentials: BigQueryCredentials) -> None:
        super().__init__(parsed_model, credentials)

        self._model_name = parsed_model["alias"]
        self._connection_method = credentials.method
        self._GoogleCredentials = create_google_credentials(credentials)
        self._notebook_client = create_notebook_client(self._GoogleCredentials, self._region)
        self._notebook_template_id = parsed_model["config"].get("notebook_template_id")
        self._packages = parsed_model["config"].get("packages", [])
        retry = RetryFactory(credentials)
        self._polling_retry = retry.create_polling(
            model_timeout=parsed_model["config"].get("timeout") or _DEFAULT_BIGFRAMES_TIMEOUT
        )

    def _get_token(self) -> str:
        """Get a token from the credentials.
        If a token is not supplied by the user directly it is lazily created the first time we authenticates.
        BigFrames needs a token to determine the execution user but a call may not have
        """
        creds = self._GoogleCredentials
        if creds.token:
            return creds.token
        else:
            creds.refresh(Request())
            return creds.token

    def _py_to_ipynb(self, compiled_code: str) -> str:
        notebook = nbformat.v4.new_notebook()
        # Put all codes in one cell.
        notebook.cells.append(nbformat.v4.new_code_cell(compiled_code))

        return nbformat.writes(notebook, nbformat.NO_CONVERT)

    def _get_notebook_template_id(self) -> str:
        # If user specifies a runtime template id, use it.
        if self._notebook_template_id:
            return self._notebook_template_id

        # Try to find and use the default runtime template id.
        request = aiplatform_v1.ListNotebookRuntimeTemplatesRequest(
            parent=f"projects/{self._project}/locations/{self._region}",
            filter="notebookRuntimeType = ONE_CLICK",
        )
        page_result = self._notebook_client.list_notebook_runtime_templates(request=request)

        try:
            # Check if a default runtime template is available and applicable.
            return self._extract_template_id(next(iter(page_result)).name)
        except Exception:
            _logger.info(
                "No default template found, a new one will be created but with "
                "disabled internet access. If your models do require internet "
                "access, please go to the GCP console and do either:\n"
                "    1. Recreate the default template yourself with enabled "
                "internet access. OR \n"
                "    2. Specify your own template ID which has enabled "
                "internet access.\n"
            )
            # If no default runtime template is found, create a new one.
            return self._create_notebook_template()

    def _create_notebook_template(self) -> str:
        # Construct the full network and subnetwork resource names.
        network_full_name = f"projects/{self._project}/global/networks/{_NETWORK_NAME}"
        subnetwork_full_name = (
            f"projects/{self._project}/regions/{self._region}/subnetworks/{_SUBNETWORK_NAME}"
        )

        template = aiplatform_v1.NotebookRuntimeTemplate(
            # The display name of the created runtime template.
            display_name="default-one-click-notebook",
            # This "ONE_CLICK" will be marked default.
            notebook_runtime_type=aiplatform_v1.NotebookRuntimeType.ONE_CLICK,
            machine_spec=aiplatform_v1.MachineSpec(
                # Choose the machine type.
                machine_type="e2-standard-4",
            ),
            network_spec=aiplatform_v1.NetworkSpec(
                # Explicitly disable internet access.
                enable_internet_access=False,
                # Need to specify the network & subnetwork (full name) when
                # disable internet access.
                network=network_full_name,
                subnetwork=subnetwork_full_name,
            ),
        )

        create_request = aiplatform_v1.CreateNotebookRuntimeTemplateRequest(
            parent=f"projects/{self._project}/locations/{self._region}",
            notebook_runtime_template=template,
        )

        operation = self._notebook_client.create_notebook_runtime_template(request=create_request)
        response = operation.result()

        return self._extract_template_id(response.name)

    def _extract_template_id(self, template_name: str) -> str:
        match = re.search(r"notebookRuntimeTemplates/(\d+)", template_name)
        return match.group(1) if match else ""

    def _config_notebook_job(
        self, notebook_template_id: str
    ) -> aiplatform_v1.NotebookExecutionJob:
        notebook_execution_job = aiplatform_v1.NotebookExecutionJob()
        notebook_execution_job.notebook_runtime_template_resource_name = (
            f"projects/{self._project}/locations/{self._region}/"
            f"notebookRuntimeTemplates/{notebook_template_id}"
        )

        notebook_execution_job.gcs_notebook_source = (
            aiplatform_v1.NotebookExecutionJob.GcsNotebookSource(uri=self._gcs_path)
        )

        if self._connection_method in (
            BigQueryConnectionMethod.SERVICE_ACCOUNT,
            BigQueryConnectionMethod.SERVICE_ACCOUNT_JSON,
        ):
            notebook_execution_job.service_account = self._GoogleCredentials._service_account_email
        elif self._connection_method in (
            BigQueryConnectionMethod.OAUTH,
            BigQueryConnectionMethod.OAUTH_SECRETS,
        ):
            # If `impersonate_service_account` is configured correctly in
            # profiles.yml, the job will run as the specified service account.
            if hasattr(self._GoogleCredentials, "_target_principal"):
                target_principal = self._GoogleCredentials._target_principal
                if target_principal:
                    notebook_execution_job.service_account = target_principal
                else:
                    raise ValueError(
                        "The impersonated service account is incorrect. Please "
                        "verify the `impersonate_service_account` setting in "
                        "your profiles.yml configuration."
                    )

            # The job will run under the identity of the authenticated user.
            else:
                request = Request()
                response = request(
                    method="GET",
                    url="https://www.googleapis.com/oauth2/v2/userinfo",
                    headers={"Authorization": f"Bearer {self._get_token()}"},
                )

                if response.status != 200:
                    raise DbtRuntimeError(
                        f"Failed to retrieve user info. Status: {response.status}, Body: {response.data}"
                    )
                if user_email := json.loads(response.data).get("email"):
                    # In services such as Cloud Composer and Cloud Run, the authenticated user
                    # is a service account with associated Application Default Credentials.
                    # This does not require service account impersonation.
                    if user_email and user_email.endswith(_SERVICE_ACCOUNT_SUFFIX):
                        notebook_execution_job.service_account = user_email
                    else:
                        notebook_execution_job.execution_user = user_email
                else:
                    raise DbtRuntimeError(
                        "Authorization request to get user failed to return an email."
                    )
        else:
            raise ValueError(
                f"Unsupported credential method in BigFrames: '{self._connection_method}'"
            )

        notebook_execution_job.gcs_output_uri = (
            f"gs://{self._gcs_bucket}/{self._model_file_name}/logs"
        )
        notebook_execution_job.display_name = self._get_batch_id()

        return notebook_execution_job

    def _read_json_from_gcs(self, gcs_uri: str) -> Optional[Any]:
        bucket_name, *blob_names = gcs_uri.split("gs://")[1].split("/")
        blob_name = "/".join(blob_names)

        bucket = self._storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)

        try:
            file_content = blob.download_as_text()
            data = json.loads(file_content)

        except json.JSONDecodeError:
            _logger.debug(f"Error: File '{gcs_uri}' content is not valid JSON.")
            data = None

        except Exception as e:
            _logger.exception(f"Error reading file from GCS: {e}")
            data = None

        return data

    def _format_outputs(self, output_list: list) -> str:
        """Formats a list of outputs into readable string."""
        formatted_output = "\n"

        # The item of it can be dictionaries or strings.
        for item in output_list:
            if isinstance(item, dict) and True:
                for key, value in item.items():
                    formatted_output += f"{key}:\n"
                    if isinstance(value, dict):
                        for inner_key, inner_value in value.items():
                            formatted_output += f"    {inner_key}: {inner_value}\n"
                    else:
                        formatted_output += f"    {value}\n"

                    # Extract only the final error message line. Discard the
                    # traceback details as they contain raw ANSI escape codes.
                    if isinstance(value, str) and value.strip().lower() == "error":
                        return formatted_output

            else:
                _logger.debug("Unexpected output format of the Colab notebook.")
                formatted_output += f"{item}\n"

            # Add a newline between items.
            formatted_output += "\n"

        return formatted_output

    def _process_gcs_log(self, gcs_log_uri: str) -> None:
        """Processes a Colab notebook execution log stored GCS."""
        gcs_log = self._read_json_from_gcs(gcs_log_uri)

        if not gcs_log:
            _logger.debug(f"Failed to read log from GCS URI: {gcs_log_uri}")
            return

        # Extract the notebook 'cells' information list from the log.
        cells = gcs_log.get("cells", [])
        if not cells:
            _logger.debug(f"No 'cells' found. Full content from GCS log: {gcs_log}")
            return

        # Only one cell exists and gets executed in the notebook.
        outputs = cells[0].get("outputs", [])
        if not outputs:
            _logger.debug(f"No 'outputs' found. Full content from GCS log: {gcs_log}")
            return

        try:
            # Improve the output format for better readability.
            formatted_output = self._format_outputs(outputs)
            _logger.info(f"Colab notebook runtime outputs from GCS: {formatted_output}")
        except Exception:
            _logger.exception(f"Failed to format the outputs from GCS: {outputs}")

    def submit(self, compiled_code: str) -> None:
        if self._packages:
            install_code = f"_install_packages({self._packages})"
            compiled_code = (
                f"{inspect.getsource(_install_packages)}\n{install_code}\n{compiled_code}"
            )
        notebook_compiled_code = self._py_to_ipynb(compiled_code)
        notebook_template_id = self._get_notebook_template_id()

        self._write_to_gcs(notebook_compiled_code)

        self._submit_bigframes_job(notebook_template_id)

    def _track_notebook_job_status(self, job_name: str) -> aiplatform_v1.NotebookExecutionJob:
        """Tracks the notebook job until it completes or times out."""
        max_wait_time = self._polling_retry.timeout
        elapsed = 0

        # Please see all the JobState from
        # https://cloud.google.com/php/docs/reference/cloud-ai-platform/latest/V1.JobState.
        terminal_states = {
            aiplatform_gapic.JobState.JOB_STATE_SUCCEEDED,
            aiplatform_gapic.JobState.JOB_STATE_PARTIALLY_SUCCEEDED,
            aiplatform_gapic.JobState.JOB_STATE_FAILED,
            aiplatform_gapic.JobState.JOB_STATE_CANCELLED,
            aiplatform_gapic.JobState.JOB_STATE_EXPIRED,
        }

        while True:
            retrieved_job = self._notebook_client.get_notebook_execution_job(name=job_name)
            job_state = retrieved_job.job_state

            if job_state in terminal_states:
                return retrieved_job

            if elapsed >= max_wait_time:
                raise TimeoutError(
                    "Operation did not complete within the designated timeout "
                    f"of {max_wait_time} seconds. Please cancel the related "
                    "notebook job manually via the GCP console since it might "
                    "still be actively running."
                )

            time.sleep(_COLAB_POLL_INTERVAL)
            elapsed += _COLAB_POLL_INTERVAL

    def _submit_bigframes_job(
        self, notebook_template_id: str
    ) -> aiplatform_v1.NotebookExecutionJob:

        notebook_execution_job = self._config_notebook_job(notebook_template_id)

        request = aiplatform_v1.CreateNotebookExecutionJobRequest(
            parent=f"projects/{self._project}/locations/{self._region}",
            notebook_execution_job=notebook_execution_job,
        )

        try:
            operation: Operation = self._notebook_client.create_notebook_execution_job(
                request=request
            )
            lro_name = operation.operation.name
            job_name = lro_name.split("/operations/")[0]
            retrieved_job = self._track_notebook_job_status(job_name)
        except TimeoutError as timeout_error:
            raise TimeoutError(f"The dbt operation encountered a timeout: {timeout_error}")
        except Exception as e:
            raise DbtRuntimeError(f"An unexpected error occured while executing the notebook: {e}")

        job_id = job_name.split("/")[-1]
        gcs_log_uri = f"{notebook_execution_job.gcs_output_uri}/{job_id}/{self._model_name}.py"
        self._process_gcs_log(gcs_log_uri)

        if retrieved_job.job_state == aiplatform_gapic.JobState.JOB_STATE_SUCCEEDED:
            _logger.info(
                f"Colab notebook execution job '{retrieved_job.name}' finished successfully."
            )
        elif retrieved_job.job_state == aiplatform_gapic.JobState.JOB_STATE_FAILED:
            raise DbtRuntimeError(
                f"The colab notebook execution job '{retrieved_job.name}' failed."
            )
        else:
            raise DbtRuntimeError(
                f"The colab notebook execution job '{retrieved_job.name}' "
                f"finished with unexpected state: {retrieved_job.job_state.name}"
            )

        return self._notebook_client.get_notebook_execution_job(name=job_name)


def _install_packages(packages: list[str]) -> None:
    """Checks and installs packages via pip in a separate environment.

    Parses requirement strings (e.g., 'pandas>=1.0', 'numpy==2.1.1') using the
    'packaging' library to check against installed versions. It only installs
    packages that are not already present in the environment. Existing packages
    will not be updated, even if a different version is requested.

    NOTE: This function is not intended for direct invocation. Instead, its
    source code is extracted using inspect.getsource for execution in a separate
    environment.
    """
    import sys
    import subprocess
    import importlib.metadata
    from packaging.requirements import Requirement
    from packaging.version import Version
    from typing import Optional, Tuple

    def _is_package_installed(requirement: Requirement) -> Tuple[bool, Optional[Version]]:
        try:
            version = importlib.metadata.version(requirement.name)
            return True, Version(version)
        except Exception:
            # Unable to determine the version.
            return False, None

    # Check the installation of individual packages first.
    packages_to_install = []
    for package in packages:
        requirement = Requirement(package)
        installed, version = _is_package_installed(requirement)

        if installed:
            if version and requirement.specifier and version not in requirement.specifier:
                print(
                    f"Package '{requirement.name}' is already installed and cannot be updated. Skipping."
                    f"The installed version: {version}."
                )
            else:
                print(
                    f"Package '{requirement.name}' is already installed. Skipping."
                    f"The installed version: {version}."
                )
        else:
            packages_to_install.append(package)

    # All packages have been installed with certain version.
    if not packages_to_install:
        return

    # Try to pip install the uninstalled packages.
    pip_command = [sys.executable, "-m", "pip", "install"] + packages_to_install
    print(f"Attempting to install the following packages: {', '.join(packages_to_install)}")

    try:
        result = subprocess.run(
            pip_command,
            capture_output=True,
            text=True,
            check=True,
            encoding="utf-8",
        )
        if result.stdout:
            print(f"pip output:\n{result.stdout.strip()}")
        if result.stderr:
            print(f"pip warnings/errors:\n{result.stderr.strip()}")
        print(f"Successfully installed the following packages: {', '.join(packages_to_install)}")

    except Exception as e:
        raise RuntimeError(f"An unexpected error occurred during package installation: {e}")



# array_append

# EXCEPT can't be used with ARRAYs in BigQuery, so convert to a string
models__array_append_expected_sql = """
select 1 as id, {{ array_to_string(array_construct([1,2,3,4])) }} as array_col union all
select 2 as id, {{ array_to_string(array_construct([4])) }} as array_col
"""


models__array_append_actual_sql = """
select 1 as id, {{ array_to_string(array_append(array_construct([1,2,3]), 4)) }} as array_col union all
select 2 as id, {{ array_to_string(array_append(array_construct([]), 4)) }} as array_col
"""



MY_SEED = """
id,value,record_date
1,100,2023-01-01 12:00:00
2,200,2023-01-02 12:00:00
3,300,2023-01-02 12:00:00
""".strip()


MY_BASE_TABLE = """
{{ config(
    materialized='table',
    partition_by={
        "field": "record_date",
        "data_type": "datetime",
        "granularity": "day"
    },
    cluster_by=["id", "value"]
) }}
select
    id,
    value,
    record_date
from {{ ref('my_seed') }}
"""


MY_MATERIALIZED_VIEW = """
{{ config(
    materialized='materialized_view',
    partition_by={
        "field": "record_date",
        "data_type": "datetime",
        "granularity": "day"
    },
    cluster_by="id",
) }}
select
    id,
    value,
    record_date
from {{ ref('my_base_table') }}
"""


MY_OTHER_BASE_TABLE = """
{{ config(
    materialized='table',
    partition_by={
        "field": "value",
        "data_type": "int64",
        "range": {
            "start": 0,
            "end": 500,
            "interval": 50
        }
    },
    cluster_by=["id", "value"]
) }}
select
    id,
    value,
    record_date
from {{ ref('my_seed') }}
"""


MY_OTHER_MATERIALIZED_VIEW = """
{{ config(
    materialized='materialized_view',
    partition_by={
        "field": "value",
        "data_type": "int64",
        "range": {
            "start": 0,
            "end": 500,
            "interval": 50
        }
    },
    cluster_by="id",
    enable_refresh=False,
    refresh_interval_minutes=60
) }}
select
    id,
    value,
    record_date
from {{ ref('my_other_base_table') }}
"""



from typing import Optional

from dbt.adapters.contracts.relation import RelationConfig

from dbt.adapters.catalogs import CATALOG_INTEGRATION_MODEL_CONFIG_NAME
from dbt.adapters.bigquery import constants


def catalog_name(model: RelationConfig) -> Optional[str]:
    # while this looks equivalent to `if not getattr(model, "config", None):`, `mypy` disagrees
    if not hasattr(model, "config") or not model.config:
        return None

    if _catalog := model.config.get(CATALOG_INTEGRATION_MODEL_CONFIG_NAME):
        return _catalog

    # TODO: deprecate this as it's been replaced with catalog_name
    if _catalog := model.config.get("catalog"):
        return _catalog

    return constants.DEFAULT_INFO_SCHEMA_CATALOG.name



"""Unit test utility functions.
Note that all imports should be inside the functions to avoid import/mocking
issues.
"""

import string
import os
from unittest import mock
from unittest import TestCase

import agate
import pytest

from dbt_common.dataclass_schema import ValidationError
from dbt.config.project import PartialProject


def normalize(path):
    """On windows, neither is enough on its own:
    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class Obj:
    which = "blah"
    single_threaded = False


def mock_connection(name, state="open"):
    conn = mock.MagicMock()
    conn.name = name
    conn.state = state
    return conn


def profile_from_dict(profile, profile_name, cli_vars="{}"):
    from dbt.config import Profile
    from dbt.config.renderer import ProfileRenderer
    from dbt.config.utils import parse_cli_vars

    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = ProfileRenderer(cli_vars)

    # in order to call dbt's internal profile rendering, we need to set the
    # flags global. This is a bit of a hack, but it's the best way to do it.
    from dbt.flags import set_from_args
    from argparse import Namespace

    set_from_args(Namespace(), None)
    return Profile.from_raw_profile_info(
        profile,
        profile_name,
        renderer,
    )


def project_from_dict(project, profile, packages=None, selectors=None, cli_vars="{}"):
    from dbt.config.renderer import DbtProjectYamlRenderer
    from dbt.config.utils import parse_cli_vars

    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = DbtProjectYamlRenderer(profile, cli_vars)

    project_root = project.pop("project-root", os.getcwd())

    partial = PartialProject.from_dicts(
        project_root=project_root,
        project_dict=project,
        packages_dict=packages,
        selectors_dict=selectors,
    )
    return partial.render(renderer)


def config_from_parts_or_dicts(project, profile, packages=None, selectors=None, cli_vars="{}"):
    from dbt.config import Project, Profile, RuntimeConfig
    from copy import deepcopy
    from dbt.config.utils import parse_cli_vars

    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    if isinstance(project, Project):
        profile_name = project.profile_name
    else:
        profile_name = project.get("profile")

    if not isinstance(profile, Profile):
        profile = profile_from_dict(
            deepcopy(profile),
            profile_name,
            cli_vars,
        )

    if not isinstance(project, Project):
        project = project_from_dict(
            deepcopy(project),
            profile,
            packages,
            selectors,
            cli_vars,
        )

    args = Obj()
    args.vars = cli_vars
    args.profile_dir = "/dev/null"
    return RuntimeConfig.from_parts(project=project, profile=profile, args=args)


def inject_plugin(plugin):
    from dbt.adapters.factory import FACTORY

    key = plugin.adapter.type()
    FACTORY.plugins[key] = plugin


def inject_plugin_for(config):
    from dbt.adapters.factory import FACTORY

    FACTORY.load_plugin(config.credentials.type)
    adapter = FACTORY.get_adapter(config)
    return adapter


def inject_adapter(value):
    """Inject the given adapter into the adapter factory, so your hand-crafted
    artisanal adapter will be available from get_adapter() as if dbt loaded it.
    """
    from dbt.adapters.factory import FACTORY

    key = value.type()
    FACTORY.adapters[key] = value


def clear_plugin(plugin):
    from dbt.adapters.factory import FACTORY

    key = plugin.adapter.type()
    FACTORY.plugins.pop(key, None)
    FACTORY.adapters.pop(key, None)


class ContractTestCase(TestCase):
    ContractType = None

    def setUp(self):
        self.maxDiff = None
        super().setUp()

    def assert_to_dict(self, obj, dct):
        self.assertEqual(obj.to_dict(omit_none=True), dct)

    def assert_from_dict(self, obj, dct, cls=None):
        if cls is None:
            cls = self.ContractType
        cls.validate(dct)
        self.assertEqual(cls.from_dict(dct), obj)

    def assert_symmetric(self, obj, dct, cls=None):
        self.assert_to_dict(obj, dct)
        self.assert_from_dict(obj, dct, cls)

    def assert_fails_validation(self, dct, cls=None):
        if cls is None:
            cls = self.ContractType

        with self.assertRaises(ValidationError):
            cls.validate(dct)
            cls.from_dict(dct)


def compare_dicts(dict1, dict2):
    first_set = set(dict1.keys())
    second_set = set(dict2.keys())
    print(f"--- Difference between first and second keys: {first_set.difference(second_set)}")
    print(f"--- Difference between second and first keys: {second_set.difference(first_set)}")
    common_keys = set(first_set).intersection(set(second_set))
    found_differences = False
    for key in common_keys:
        if dict1[key] != dict2[key]:
            print(f"--- --- first dict: {key}: {str(dict1[key])}")
            print(f"--- --- second dict: {key}: {str(dict2[key])}")
            found_differences = True
    if found_differences:
        print("--- Found differences in dictionaries")
    else:
        print("--- Found no differences in dictionaries")


def assert_from_dict(obj, dct, cls=None):
    if cls is None:
        cls = obj.__class__
    cls.validate(dct)
    obj_from_dict = cls.from_dict(dct)
    if hasattr(obj, "created_at"):
        obj_from_dict.created_at = 1
        obj.created_at = 1
    assert obj_from_dict == obj


def assert_to_dict(obj, dct):
    obj_to_dict = obj.to_dict(omit_none=True)
    if "created_at" in obj_to_dict:
        obj_to_dict["created_at"] = 1
    if "created_at" in dct:
        dct["created_at"] = 1
    assert obj_to_dict == dct


def assert_symmetric(obj, dct, cls=None):
    assert_to_dict(obj, dct)
    assert_from_dict(obj, dct, cls)


def assert_fails_validation(dct, cls):
    with pytest.raises(ValidationError):
        cls.validate(dct)
        cls.from_dict(dct)


def generate_name_macros(package):
    from dbt.contracts.graph.nodes import Macro
    from dbt.node_types import NodeType

    name_sql = {}
    for component in ("database", "schema", "alias"):
        if component == "alias":
            source = "node.name"
        else:
            source = f"target.{component}"
        name = f"generate_{component}_name"
        sql = f"{{% macro {name}(value, node) %}} {{% if value %}} {{{{ value }}}} {{% else %}} {{{{ {source} }}}} {{% endif %}} {{% endmacro %}}"
        name_sql[name] = sql

    for name, sql in name_sql.items():
        pm = Macro(
            name=name,
            resource_type=NodeType.Macro,
            unique_id=f"macro.{package}.{name}",
            package_name=package,
            original_file_path=normalize("macros/macro.sql"),
            path=normalize("macros/macro.sql"),
            macro_sql=sql,
        )
        yield pm


class TestAdapterConversions(TestCase):
    def _get_tester_for(self, column_type):
        from dbt_common.clients import agate_helper

        if column_type is agate.TimeDelta:  # dbt never makes this!
            return agate.TimeDelta()

        for instance in agate_helper.DEFAULT_TYPE_TESTER._possible_types:
            if isinstance(instance, column_type):  # include child types
                return instance

        raise ValueError(f"no tester for {column_type}")

    def _make_table_of(self, rows, column_types):
        column_names = list(string.ascii_letters[: len(rows[0])])
        if isinstance(column_types, type):
            column_types = [self._get_tester_for(column_types) for _ in column_names]
        else:
            column_types = [self._get_tester_for(typ) for typ in column_types]
        table = agate.Table(rows, column_names=column_names, column_types=column_types)
        return table


def MockMacro(package, name="my_macro", **kwargs):
    from dbt.contracts.graph.nodes import Macro
    from dbt.node_types import NodeType

    mock_kwargs = dict(
        resource_type=NodeType.Macro,
        package_name=package,
        unique_id=f"macro.{package}.{name}",
        original_file_path="/dev/null",
    )

    mock_kwargs.update(kwargs)

    macro = mock.MagicMock(spec=Macro, **mock_kwargs)
    macro.name = name
    return macro


def MockMaterialization(package, name="my_materialization", adapter_type=None, **kwargs):
    if adapter_type is None:
        adapter_type = "default"
    kwargs["adapter_type"] = adapter_type
    return MockMacro(package, f"materialization_{name}_{adapter_type}", **kwargs)


def MockGenerateMacro(package, component="some_component", **kwargs):
    name = f"generate_{component}_name"
    return MockMacro(package, name=name, **kwargs)


def MockSource(package, source_name, name, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.nodes import SourceDefinition

    src = mock.MagicMock(
        __class__=SourceDefinition,
        resource_type=NodeType.Source,
        source_name=source_name,
        package_name=package,
        unique_id=f"source.{package}.{source_name}.{name}",
        search_name=f"{source_name}.{name}",
        **kwargs,
    )
    src.name = name
    return src


def MockNode(package, name, resource_type=None, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.nodes import ModelNode, SeedNode

    if resource_type is None:
        resource_type = NodeType.Model
    if resource_type == NodeType.Model:
        cls = ModelNode
    elif resource_type == NodeType.Seed:
        cls = SeedNode
    else:
        raise ValueError(f"I do not know how to handle {resource_type}")
    node = mock.MagicMock(
        __class__=cls,
        resource_type=resource_type,
        package_name=package,
        unique_id=f"{str(resource_type)}.{package}.{name}",
        search_name=name,
        **kwargs,
    )
    node.name = name
    return node


def MockDocumentation(package, name, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.nodes import Documentation

    doc = mock.MagicMock(
        __class__=Documentation,
        resource_type=NodeType.Documentation,
        package_name=package,
        search_name=name,
        unique_id=f"{package}.{name}",
        **kwargs,
    )
    doc.name = name
    return doc


def load_internal_manifest_macros(config, macro_hook=lambda m: None):
    from dbt.parser.manifest import ManifestLoader

    return ManifestLoader.load_macros(config, macro_hook)


def dict_replace(dct, **kwargs):
    dct = dct.copy()
    dct.update(kwargs)
    return dct


def replace_config(n, **kwargs):
    return n.replace(
        config=n.config.replace(**kwargs),
        unrendered_config=dict_replace(n.unrendered_config, **kwargs),
    )



import random

import pytest
from google.api_core.exceptions import NotFound

from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct
from dbt.tests.adapter.utils.test_any_value import BaseAnyValue
from dbt.tests.adapter.utils.test_bool_or import BaseBoolOr
from dbt.tests.adapter.utils.test_cast import BaseCast
from dbt.tests.adapter.utils.test_cast_bool_to_text import BaseCastBoolToText
from dbt.tests.adapter.utils.test_concat import BaseConcat
from dbt.tests.adapter.utils.test_current_timestamp import BaseCurrentTimestampAware
from dbt.tests.adapter.utils.test_date import BaseDate
from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd
from dbt.tests.adapter.utils.test_datediff import BaseDateDiff
from dbt.tests.adapter.utils.test_date_spine import BaseDateSpine
from dbt.tests.adapter.utils.test_date_trunc import BaseDateTrunc
from dbt.tests.adapter.utils.test_escape_single_quotes import BaseEscapeSingleQuotesBackslash
from dbt.tests.adapter.utils.test_except import BaseExcept
from dbt.tests.adapter.utils.test_generate_series import BaseGenerateSeries
from dbt.tests.adapter.utils.test_get_intervals_between import BaseGetIntervalsBetween
from dbt.tests.adapter.utils.test_get_powers_of_two import BaseGetPowersOfTwo
from dbt.tests.adapter.utils.test_hash import BaseHash
from dbt.tests.adapter.utils.test_intersect import BaseIntersect
from dbt.tests.adapter.utils.test_last_day import BaseLastDay
from dbt.tests.adapter.utils.test_length import BaseLength
from dbt.tests.adapter.utils.test_listagg import BaseListagg
from dbt.tests.adapter.utils.test_position import BasePosition
from dbt.tests.adapter.utils.test_replace import BaseReplace
from dbt.tests.adapter.utils.test_right import BaseRight
from dbt.tests.adapter.utils.test_safe_cast import BaseSafeCast
from dbt.tests.adapter.utils.test_split_part import BaseSplitPart
from dbt.tests.adapter.utils.test_string_literal import BaseStringLiteral
from dbt.tests.adapter.utils.test_validate_sql import BaseValidateSqlMethod
from tests.functional.adapter.utils.fixture_array_append import (
    models__array_append_actual_sql,
    models__array_append_expected_sql,
)
from tests.functional.adapter.utils.fixture_array_concat import (
    models__array_concat_actual_sql,
    models__array_concat_expected_sql,
)
from tests.functional.adapter.utils.fixture_array_construct import (
    models__array_construct_actual_sql,
    models__array_construct_expected_sql,
    macros__array_to_string_sql,
)
from tests.functional.adapter.utils.fixture_get_intervals_between import (
    models__bq_test_get_intervals_between_sql,
    models___bq_test_get_intervals_between_yml,
)


class TestAnyValue(BaseAnyValue):
    pass


class TestArrayAppend(BaseArrayAppend):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "actual.sql": models__array_append_actual_sql,
            "expected.sql": models__array_append_expected_sql,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "array_to_string.sql": macros__array_to_string_sql,
        }


class TestArrayConcat(BaseArrayConcat):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "actual.sql": models__array_concat_actual_sql,
            "expected.sql": models__array_concat_expected_sql,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "array_to_string.sql": macros__array_to_string_sql,
        }


class TestArrayConstruct(BaseArrayConstruct):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "actual.sql": models__array_construct_actual_sql,
            "expected.sql": models__array_construct_expected_sql,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "array_to_string.sql": macros__array_to_string_sql,
        }


class TestBoolOr(BaseBoolOr):
    pass


class TestCast(BaseCast):
    pass


class TestCastBoolToText(BaseCastBoolToText):
    pass


class TestConcat(BaseConcat):
    pass


# Use either BaseCurrentTimestampAware or BaseCurrentTimestampNaive but not both
class TestCurrentTimestamp(BaseCurrentTimestampAware):
    pass


class TestDate(BaseDate):
    pass


class TestDateAdd(BaseDateAdd):
    pass


class TestDateDiff(BaseDateDiff):
    pass


class TestDateSpine(BaseDateSpine):
    pass


class TestDateTrunc(BaseDateTrunc):
    pass


class TestEscapeSingleQuotes(BaseEscapeSingleQuotesBackslash):
    pass


class TestExcept(BaseExcept):
    pass


class TestGenerateSeries(BaseGenerateSeries):
    pass


class TestGetIntervalsBetween(BaseGetIntervalsBetween):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "test_get_intervals_between.yml": models___bq_test_get_intervals_between_yml,
            "test_get_intervals_between.sql": self.interpolate_macro_namespace(
                models__bq_test_get_intervals_between_sql, "get_intervals_between"
            ),
        }


class TestGetPowersOfTwo(BaseGetPowersOfTwo):
    pass


class TestHash(BaseHash):
    pass


class TestIntersect(BaseIntersect):
    pass


class TestLastDay(BaseLastDay):
    pass


class TestLength(BaseLength):
    pass


class TestListagg(BaseListagg):
    pass


class TestPosition(BasePosition):
    pass


class TestReplace(BaseReplace):
    pass


class TestRight(BaseRight):
    pass


class TestSafeCast(BaseSafeCast):
    pass


class TestSplitPart(BaseSplitPart):
    pass


class TestStringLiteral(BaseStringLiteral):
    pass


class TestValidateSqlMethod(BaseValidateSqlMethod):
    pass


class TestDryRunMethod:
    """Test connection manager dry run method operation."""

    def test_dry_run_method(self, project) -> None:
        """Test dry run method on a DDL statement.

        This allows us to demonstrate that no SQL is executed.
        """
        with project.adapter.connection_named("_test"):
            client = project.adapter.connections.get_thread_connection().handle
            random_suffix = "".join(random.choices([str(i) for i in range(10)], k=10))
            table_name = f"test_dry_run_{random_suffix}"
            table_id = "{}.{}.{}".format(project.database, project.test_schema, table_name)
            res = project.adapter.connections.dry_run(f"CREATE TABLE {table_id} (x INT64)")
            assert res.code == "DRY RUN"
            with pytest.raises(expected_exception=NotFound):
                client.get_table(table_id)



import pytest
from dbt.tests.util import run_dbt, get_connection, relation_from_name

from dbt.adapters.bigquery import BigQueryRelation

_POLICY_TAG_MODEL = """{{
  config(
    materialized='table',
    persist_docs={ 'columns': true }
  )
}}

select
  1 field
"""

_POLICY_TAG_YML = """version: 2

models:
- name: policy_tag_table
  columns:
  - name: field
    policy_tags:
      - '{{ var("policy_tag") }}'
"""

# Manually generated https://console.cloud.google.com/bigquery/policy-tags?project=dbt-test-env
_POLICY_TAG = "projects/dbt-test-env/locations/us/taxonomies/5785568062805976401/policyTags/135489647357012267"
_POLICY_TAG_MODEL_NAME = "policy_tag_table"


class TestBigqueryUpdateColumnPolicy:
    """See BQ docs for more info on policy tags:
    https://cloud.google.com/bigquery/docs/column-level-security#work_with_policy_tags"""

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"config-version": 2, "vars": {"policy_tag": _POLICY_TAG}}

    @pytest.fixture(scope="class")
    def models(self):
        return {f"{_POLICY_TAG_MODEL_NAME}.sql": _POLICY_TAG_MODEL, "schema.yml": _POLICY_TAG_YML}

    def test_bigquery_update_column_policy_tag(self, project):
        results = run_dbt(["run", "--models", "policy_tag_table"])
        assert len(results) == 1
        relation: BigQueryRelation = relation_from_name(project.adapter, _POLICY_TAG_MODEL_NAME)
        adapter = project.adapter
        with get_connection(project.adapter) as conn:
            table = conn.handle.get_table(
                adapter.connections.get_bq_table(
                    relation.database, relation.schema, relation.table
                )
            )
            for schema_field in table.schema:
                assert schema_field.policy_tags.names == (_POLICY_TAG,)



import pytest
from dbt.tests.adapter.constraints.test_constraints import (
    BaseTableConstraintsColumnsEqual,
    BaseViewConstraintsColumnsEqual,
    BaseIncrementalConstraintsColumnsEqual,
    BaseTableContractSqlHeader,
    BaseIncrementalContractSqlHeader,
    BaseConstraintsRuntimeDdlEnforcement,
    BaseConstraintsRollback,
    BaseIncrementalConstraintsRuntimeDdlEnforcement,
    BaseIncrementalConstraintsRollback,
    BaseModelConstraintsRuntimeEnforcement,
    BaseConstraintQuotedColumn,
)
from dbt.tests.adapter.constraints.fixtures import (
    my_model_sql,
    my_incremental_model_sql,
    my_model_wrong_order_sql,
    my_model_view_wrong_order_sql,
    my_model_incremental_wrong_order_sql,
    my_model_wrong_name_sql,
    my_model_view_wrong_name_sql,
    my_model_incremental_wrong_name_sql,
    my_model_with_quoted_column_name_sql,
    model_schema_yml,
    constrained_model_schema_yml,
    model_contract_header_schema_yml,
    model_quoted_column_schema_yml,
    model_fk_constraint_schema_yml,
    my_model_wrong_order_depends_on_fk_sql,
    foreign_key_model_sql,
    my_model_incremental_wrong_order_depends_on_fk_sql,
)

from tests.functional.adapter.constraints.fixtures import (
    my_model_struct_wrong_data_type_sql,
    my_model_struct_correct_data_type_sql,
    my_model_double_struct_wrong_data_type_sql,
    my_model_double_struct_correct_data_type_sql,
    model_struct_data_type_schema_yml,
    model_double_struct_data_type_schema_yml,
    my_model_struct_sql,
    model_struct_schema_yml,
)

from dbt.tests.util import run_dbt_and_capture, run_dbt

_expected_sql_bigquery = """
create or replace table <model_identifier> (
    id INT64 not null primary key not enforced references <foreign_key_model_identifier> (id) not enforced,
    color string,
    date_day string
)
OPTIONS()
as (
    select id,
    color,
    date_day from
  (
    -- depends_on: <foreign_key_model_identifier>
    select 'blue' as color,
    1 as id,
    '2019-01-01' as date_day
  ) as model_subq
);
"""

_expected_struct_sql_bigquery = """
create or replace table <model_identifier> (
    id struct<nested_column string not null, nested_column2 string>
)
OPTIONS()
as (
    select id from
  (
    select STRUCT("test" as nested_column, "test" as nested_column2) as id
  ) as model_subq
);
"""

# Different on BigQuery:
# Switch from text to string handled by aliasing
constraints_yml = model_schema_yml
model_constraints_yml = constrained_model_schema_yml

my_model_contract_sql_header_sql = """
{{
  config(
    materialized = "table"
  )
}}

{% call set_sql_header(config) %}
DECLARE DEMO STRING DEFAULT 'hello world';
{% endcall %}

SELECT DEMO as column_name
"""

my_model_incremental_contract_sql_header_sql = """
{{
  config(
    materialized = "incremental",
    on_schema_change="append_new_columns"
  )
}}

{% call set_sql_header(config) %}
DECLARE DEMO STRING DEFAULT 'hello world';
{% endcall %}

SELECT DEMO as column_name
"""


class BigQueryColumnEqualSetup:
    @pytest.fixture
    def string_type(self):
        return "STRING"

    @pytest.fixture
    def int_type(self):
        return "INT64"

    @pytest.fixture
    def data_types(self, int_type, string_type):
        # sql_column_value, schema_data_type, error_data_type
        return [
            ["1", int_type, int_type],
            ["'1'", string_type, string_type],
            ["cast('2019-01-01' as date)", "date", "DATE"],
            ["true", "bool", "BOOL"],
            ["cast('2013-11-03 00:00:00-07' as TIMESTAMP)", "timestamp", "TIMESTAMP"],
            ["['a','b','c']", f"ARRAY<{string_type}>", f"ARRAY<{string_type}>"],
            ["[1,2,3]", f"ARRAY<{int_type}>", f"ARRAY<{int_type}>"],
            ["cast(1 as NUMERIC)", "numeric", "NUMERIC"],
            ["""JSON '{"name": "Cooper", "forname": "Alice"}'""", "json", "JSON"],
        ]


class TestBigQueryTableConstraintsColumnsEqual(
    BigQueryColumnEqualSetup, BaseTableConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


class TestBigQueryViewConstraintsColumnsEqual(
    BigQueryColumnEqualSetup, BaseViewConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_view_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_view_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


class TestBigQueryIncrementalConstraintsColumnsEqual(
    BigQueryColumnEqualSetup, BaseIncrementalConstraintsColumnsEqual
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_wrong_order.sql": my_model_incremental_wrong_order_sql,
            "my_model_wrong_name.sql": my_model_incremental_wrong_name_sql,
            "constraints_schema.yml": constraints_yml,
        }


class TestBigQueryTableContractsSqlHeader(BaseTableContractSqlHeader):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_contract_sql_header.sql": my_model_contract_sql_header_sql,
            "constraints_schema.yml": model_contract_header_schema_yml,
        }


class TestBigQueryIncrementalContractsSqlHeader(BaseIncrementalContractSqlHeader):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model_contract_sql_header.sql": my_model_incremental_contract_sql_header_sql,
            "constraints_schema.yml": model_contract_header_schema_yml,
        }


class BaseStructContract:
    @pytest.fixture
    def wrong_schema_data_type(self):
        return "INT64"

    @pytest.fixture
    def correct_schema_data_type(self):
        return "STRING"

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "contract_struct_schema.yml": model_struct_data_type_schema_yml,
            "contract_struct_wrong.sql": my_model_struct_wrong_data_type_sql,
            "contract_struct_correct.sql": my_model_struct_correct_data_type_sql,
        }

    def test__struct_contract_wrong_data_type(
        self, project, correct_schema_data_type, wrong_schema_data_type
    ):
        results, log_output = run_dbt_and_capture(
            ["run", "-s", "contract_struct_wrong"], expect_pass=False
        )
        assert len(results) == 1
        assert results[0].node.config.contract.enforced

        expected = [
            "struct_column_being_tested",
            wrong_schema_data_type,
            correct_schema_data_type,
            "data type mismatch",
        ]
        assert all([(exp in log_output or exp.upper() in log_output) for exp in expected])

    def test__struct_contract_correct_data_type(self, project):
        results = run_dbt(["run", "-s", "contract_struct_correct"])

        assert len(results) == 1
        assert results[0].node.config.contract.enforced


class TestBigQueryStructContract(BaseStructContract):
    pass


class TestBigQueryDoubleStructContract(BaseStructContract):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "contract_struct_schema.yml": model_double_struct_data_type_schema_yml,
            "contract_struct_wrong.sql": my_model_double_struct_wrong_data_type_sql,
            "contract_struct_correct.sql": my_model_double_struct_correct_data_type_sql,
        }


class TestBigQueryTableConstraintsRuntimeDdlEnforcement(BaseConstraintsRuntimeDdlEnforcement):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": model_fk_constraint_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self, project):
        return _expected_sql_bigquery


class TestBigQueryStructTableConstraintsRuntimeDdlEnforcement(
    BaseConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_struct_sql,
            "constraints_schema.yml": model_struct_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self, project):
        return _expected_struct_sql_bigquery


class TestBigQueryTableConstraintsRollback(BaseConstraintsRollback):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_sql,
            "constraints_schema.yml": constraints_yml,
        }

    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["Required field id cannot be null"]


class TestBigQueryIncrementalConstraintsRuntimeDdlEnforcement(
    BaseIncrementalConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_incremental_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": model_fk_constraint_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self, project):
        return _expected_sql_bigquery


class TestBigQueryIncrementalConstraintsRollback(BaseIncrementalConstraintsRollback):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_incremental_model_sql,
            "constraints_schema.yml": constraints_yml,
        }

    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["Required field id cannot be null"]


class TestBigQueryModelConstraintsRuntimeEnforcement(BaseModelConstraintsRuntimeEnforcement):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_wrong_order_depends_on_fk_sql,
            "foreign_key_model.sql": foreign_key_model_sql,
            "constraints_schema.yml": constrained_model_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
create or replace table <model_identifier> (
    id INT64 not null,
    color string,
    date_day string,
    primary key (id) not enforced,
    foreign key (id) references <foreign_key_model_identifier> (id) not enforced
)
OPTIONS()
as (
    select id,
    color,
    date_day from
  (
    -- depends_on: <foreign_key_model_identifier>
    select
    'blue' as color,
    1 as id,
    '2019-01-01' as date_day
  ) as model_subq
);
"""


class TestBigQueryConstraintQuotedColumn(BaseConstraintQuotedColumn):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "my_model.sql": my_model_with_quoted_column_name_sql,
            "constraints_schema.yml": model_quoted_column_schema_yml,
        }

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """
create or replace table <model_identifier> (
    id INT64 not null,
    `from` string not null,
    date_day string
)
options()
as (
    select id, `from`, date_day
    from (
        select
          'blue' as `from`,
          1 as id,
          '2019-01-01' as date_day
    ) as model_subq
);
"""



from dbt.tests.adapter.relations.test_changing_relation_type import BaseChangeRelationTypeValidator


class TestBigQueryChangeRelationTypes(BaseChangeRelationTypeValidator):
    def test_changing_materialization_changes_relation_type(self, project):
        self._run_and_check_materialization("view")
        self._run_and_check_materialization("table", extra_args=["--full-refresh"])
        self._run_and_check_materialization("view", extra_args=["--full-refresh"])
        self._run_and_check_materialization("incremental", extra_args=["--full-refresh"])



import base64
import json
import pytest
from dbt.adapters.bigquery.credentials import _is_base64


def string_to_base64(s):
    return base64.b64encode(s.encode("utf-8"))


@pytest.fixture
def example_json_keyfile():
    keyfile = json.dumps(
        {
            "type": "service_account",
            "project_id": "",
            "private_key_id": "",
            "private_key": "-----BEGIN PRIVATE KEY----------END PRIVATE KEY-----\n",
            "client_email": "",
            "client_id": "",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "",
        }
    )

    return keyfile


@pytest.fixture
def example_json_keyfile_b64():
    keyfile = json.dumps(
        {
            "type": "service_account",
            "project_id": "",
            "private_key_id": "",
            "private_key": "-----BEGIN PRIVATE KEY----------END PRIVATE KEY-----\n",
            "client_email": "",
            "client_id": "",
            "auth_uri": "https://accounts.google.com/o/oauth2/auth",
            "token_uri": "https://oauth2.googleapis.com/token",
            "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
            "client_x509_cert_url": "",
        }
    )

    return string_to_base64(keyfile)


def test_valid_base64_strings(example_json_keyfile_b64):
    valid_strings = [
        "SGVsbG8gV29ybGQh",  # "Hello World!"
        "Zm9vYmFy",  # "foobar"
        "QUJDREVGR0hJSktMTU5PUFFSU1RVVldYWVowMTIzNDU2Nzg5",  # A long string
        "",  # Empty string
        example_json_keyfile_b64.decode("utf-8"),
    ]

    for s in valid_strings:
        assert _is_base64(s) is True


def test_valid_base64_bytes(example_json_keyfile_b64):
    valid_bytes = [
        b"SGVsbG8gV29ybGQh",  # "Hello World!"
        b"Zm9vYmFy",  # "foobar"
        b"QUJDREVGR0hJSktMTU5PUFFSU1RVVldYWVowMTIzNDU2Nzg5",  # A long string
        b"",  # Empty bytes
        example_json_keyfile_b64,
    ]
    for s in valid_bytes:
        assert _is_base64(s) is True


def test_invalid_base64(example_json_keyfile):
    invalid_inputs = [
        "This is not Base64",
        "SGVsbG8gV29ybGQ",  # Incorrect padding
        "Invalid#Base64",
        12345,  # Not a string or bytes
        b"Invalid#Base64",
        "H\xffGVsbG8gV29ybGQh",  # Contains invalid character \xff
        example_json_keyfile,
    ]
    for s in invalid_inputs:
        assert _is_base64(s) is False



import os

from tests.functional.utils import unique_prefix


# make sure this is static
PREFIX = unique_prefix()


SEED = """
id,value
1,100
2,200
3,300
""".strip()


VIEW = """
{{ config(
    materialized='view',
) }}
select * from {{ ref('my_seed') }}
"""


TABLE = """
{{ config(
    materialized='table',
) }}
select * from {{ ref('my_seed') }}
"""


MATERIALIZED_VIEW = """
{{ config(
    materialized='materialized_view',
) }}
select * from {{ ref('my_seed') }}
"""


INCREMENTAL_TABLE = """
{{ config(
    materialized='incremental',
    unique_key="id",
) }}
select * from {{ ref('my_seed') }}
"""


ICEBERG_TABLE = (
    """
{{ config(
    materialized='table',
    catalog='managed_iceberg',
    storage_uri='gs://"""
    + os.getenv("BIGQUERY_TEST_ICEBERG_BUCKET")
    + """/"""
    + PREFIX
    + """__||storage_uri||'
) }}
select * from {{ ref('my_seed') }}
"""
)


INCREMENTAL_ICEBERG_TABLE = (
    """
{{ config(
    materialized='incremental',
    unique_key="id",
    catalog='managed_iceberg',
    storage_uri='gs://"""
    + os.getenv("BIGQUERY_TEST_ICEBERG_BUCKET")
    + """/"""
    + PREFIX
    + """__||storage_uri||'
) }}
select * from {{ ref('my_seed') }}
"""
)



import pytest
from dbt.tests.adapter.simple_seed.fixtures import macros__schema_test
from dbt.tests.adapter.simple_seed.seeds import seeds__enabled_in_config_csv, seeds__tricky_csv
from dbt.tests.adapter.simple_seed.test_seed import SeedConfigBase
from dbt.tests.adapter.simple_seed.test_seed import BaseTestEmptySeed
from dbt.tests.adapter.utils.base_utils import run_dbt

_SEED_CONFIGS_CSV = """
seed_id,stuff
1,a
2,b
""".lstrip()

_SCHEMA_YML = """
version: 2
seeds:
- name: seed_enabled
  columns:
  - name: birthday
    tests:
    - column_type:
        type: STRING
  - name: seed_id
    tests:
    - column_type:
        type: FLOAT64

- name: seed_tricky
  columns:
  - name: seed_id
    tests:
    - column_type:
        type: INT64
  - name: seed_id_str
    tests:
    - column_type:
        type: STRING
  - name: a_bool
    tests:
    - column_type:
        type: BOOLEAN
  - name: looks_like_a_bool
    tests:
    - column_type:
        type: STRING
  - name: a_date
    tests:
    - column_type:
        type: DATETIME
  - name: looks_like_a_date
    tests:
    - column_type:
        type: STRING
  - name: relative
    tests:
    - column_type:
        type: STRING
  - name: weekday
    tests:
    - column_type:
        type: STRING

- name: seed_configs
  config:
    hours_to_expiration: 2
    labels:
      contains_pii: 'yes'
      contains_pie: 'no'
""".lstrip()


class TestSimpleSeedConfigs(SeedConfigBase):
    @pytest.fixture(scope="class")
    def schema(self):
        return "simple_seed"

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed_enabled.csv": seeds__enabled_in_config_csv,
            "seed_tricky.csv": seeds__tricky_csv,
            "seed_configs.csv": _SEED_CONFIGS_CSV,
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "schema_test.sql": macros__schema_test,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {"models-bq.yml": _SCHEMA_YML}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "config-version": 2,
            "seeds": {
                "test": {
                    "enabled": False,
                    "quote_columns": True,
                    "seed_enabled": {
                        "enabled": True,
                        "+column_types": self.seed_enabled_types(),
                    },
                    "seed_tricky": {
                        "enabled": True,
                        "+column_types": self.seed_tricky_types(),
                    },
                    "seed_configs": {
                        "enabled": True,
                    },
                },
            },
        }

    @staticmethod
    def seed_enabled_types():
        return {
            "seed_id": "FLOAT64",
            "birthday": "STRING",
        }

    @staticmethod
    def seed_tricky_types():
        return {
            "seed_id_str": "STRING",
            "looks_like_a_bool": "STRING",
            "looks_like_a_date": "STRING",
        }

    @staticmethod
    def table_labels():
        return {"contains_pii": "yes", "contains_pie": "no"}

    def test__bigquery_simple_seed_with_column_override_bigquery(self, project):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 3
        test_results = run_dbt(["test"])
        assert len(test_results) == 10

    def test__bigquery_seed_table_with_labels_config_bigquery(self, project):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 3
        with project.adapter.connection_named("_test"):
            client = project.adapter.connections.get_thread_connection().handle
            table_id = "{}.{}.{}".format(project.database, project.test_schema, "seed_configs")
            bq_table = client.get_table(table_id)

            assert bq_table.labels
            assert bq_table.labels == self.table_labels()
            assert bq_table.expires


class TestBigQueryEmptySeed(BaseTestEmptySeed):
    pass


class TestBigQuerySeedWithUniqueDelimiter(TestSimpleSeedConfigs):
    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed_enabled.csv": seeds__enabled_in_config_csv.replace(",", "|"),
            "seed_tricky.csv": seeds__tricky_csv.replace(",", "\t"),
            "seed_configs.csv": _SEED_CONFIGS_CSV,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "config-version": 2,
            "seeds": {
                "test": {
                    "enabled": False,
                    "quote_columns": True,
                    "seed_enabled": {
                        "enabled": True,
                        "+column_types": self.seed_enabled_types(),
                        "delimiter": "|",
                    },
                    "seed_tricky": {
                        "enabled": True,
                        "+column_types": self.seed_tricky_types(),
                        "delimiter": "\t",
                    },
                    "seed_configs": {
                        "enabled": True,
                    },
                },
            },
        }



import os
from datetime import datetime as dt
import pytest
from dbt.tests.adapter.catalog_integrations.test_catalog_integration import (
    BaseCatalogIntegrationValidation,
)
from dbt.tests.util import run_dbt

_BQ_BUCKET = f"gs://{os.getenv('BIGQUERY_TEST_ICEBERG_BUCKET')}"
_STATIC_URI = f"{_BQ_BUCKET}/{str(dt.now())}"

MODEL__BASIC_ICEBERG_TABLE = """
                            {{ config(materialized='table', catalog='basic_iceberg_catalog') }}
                            select 1 as id
                            """

MODEL__SPECIFY_LOCATION_TABLE = """
                            {{ config(materialized='table', catalog='basic_iceberg_catalog',
                            base_location_root='custom_location') }}
                            select 1 as id
                            """

MODEL__SPECIFY_URI_TABLE = (
    """
                            {{ config(materialized='table', catalog='basic_iceberg_catalog',
                            storage_uri='"""
    + _STATIC_URI
    + """') }}
                            select 1 as id
                            """
)


class TestGenericCatalogIntegration(BaseCatalogIntegrationValidation):

    @pytest.fixture(scope="class")
    def catalogs(self):
        return {
            "catalogs": [
                {
                    "name": "basic_iceberg_catalog",
                    "active_write_integration": "basic_iceberg_catalog_integration",
                    "write_integrations": [
                        {
                            "name": "basic_iceberg_catalog_integration",
                            "catalog_type": "biglake_metastore",
                            "file_format": "parquet",
                            "table_format": "iceberg",
                            "external_volume": _BQ_BUCKET,
                        }
                    ],
                },
            ]
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "models": {
                "basic_iceberg_table.sql": MODEL__BASIC_ICEBERG_TABLE,
                "specify_location_table.sql": MODEL__SPECIFY_LOCATION_TABLE,
                "specify_uri_table.sql": MODEL__SPECIFY_URI_TABLE,
            }
        }

    def test_basic_iceberg_catalog_integration(self, project):
        run_dbt(["run"])



from dbt.adapters.catalogs import CatalogIntegration
from dbt.adapters.contracts.relation import RelationConfig

from dbt.adapters.bigquery import constants
from dbt.adapters.bigquery.catalogs._relation import BigQueryCatalogRelation


class BigQueryInfoSchemaCatalogIntegration(CatalogIntegration):
    catalog_type = constants.DEFAULT_INFO_SCHEMA_CATALOG.catalog_type
    allows_writes = True

    def build_relation(self, model: RelationConfig) -> BigQueryCatalogRelation:
        """
        Args:
            model: `config.model` (not `model`) from the jinja context
        """

        return BigQueryCatalogRelation(
            catalog_type=self.catalog_type,
            catalog_name=self.catalog_name,
            table_format=self.table_format,
            file_format=self.file_format,
            external_volume=None,
            storage_uri=None,
        )



import base64
import binascii
from dataclasses import dataclass, field
from functools import lru_cache
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from google.auth import default
from google.auth.exceptions import DefaultCredentialsError
from google.auth.impersonated_credentials import Credentials as ImpersonatedCredentials
from google.oauth2.credentials import Credentials as GoogleCredentials
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from google.auth.identity_pool import Credentials as IdentityPoolCredentials
from mashumaro import pass_through

from dbt_common.clients.system import run_cmd
from dbt_common.dataclass_schema import ExtensibleDbtClassMixin, StrEnum
from dbt_common.exceptions import DbtConfigError, DbtRuntimeError
from dbt.adapters.contracts.connection import Credentials
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.exceptions.connection import FailedToConnectError
from dbt.adapters.bigquery.token_suppliers import create_token_supplier


_logger = AdapterLogger("BigQuery")


class Priority(StrEnum):
    Interactive = "interactive"
    Batch = "batch"


@dataclass
class DataprocBatchConfig(ExtensibleDbtClassMixin):
    def __init__(self, batch_config):
        self.batch_config = batch_config


class BigQueryConnectionMethod(StrEnum):
    OAUTH = "oauth"
    OAUTH_SECRETS = "oauth-secrets"
    SERVICE_ACCOUNT = "service-account"
    SERVICE_ACCOUNT_JSON = "service-account-json"
    # WIF in this context refers to Workload Identity Federation https://cloud.google.com/iam/docs/workload-identity-federation
    EXTERNAL_OAUTH_WIF = "external-oauth-wif"


@dataclass
class BigQueryCredentials(Credentials):
    method: BigQueryConnectionMethod = None  # type: ignore

    # BigQuery allows an empty database / project, where it defers to the
    # environment for the project
    database: Optional[str] = None  # type:ignore
    schema: Optional[str] = None  # type:ignore
    execution_project: Optional[str] = None
    quota_project: Optional[str] = None
    location: Optional[str] = None
    api_endpoint: Optional[str] = None
    priority: Optional[Priority] = None
    maximum_bytes_billed: Optional[int] = None
    impersonate_service_account: Optional[str] = None

    job_retry_deadline_seconds: Optional[int] = None
    job_retries: Optional[int] = 1
    job_creation_timeout_seconds: Optional[int] = None
    job_execution_timeout_seconds: Optional[int] = None

    # Keyfile json creds (unicode or base 64 encoded)
    keyfile: Optional[str] = None
    keyfile_json: Optional[Dict[str, Any]] = None

    # oauth-secrets
    token: Optional[str] = None
    refresh_token: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None
    token_uri: Optional[str] = None

    # workload identity federation

    # workload_pool_provider_path
    #  The Security Token Service audience, which is usually the fully specified resource name of the workload pool provider.
    #  This field is equivalent to the `audience` key in an Application Default Credentials file downloaded from Google Cloud Platform.
    #  ex: //iam.googleapis.com/projects/<project-id>/locations/global/workloadIdentityPools/<workload-identity-pool>/providers/<workload-identity-provider>
    workload_pool_provider_path: Optional[str] = None
    # service_account_impersonation_url
    #   The URL for the service account impersonation request, used to generate access tokens via GCP's IAM service.
    #   ex: https://iamcredentials.googleapis.com/v1/projects/-/serviceAccounts/<service-account-email>:generateAccessToken
    service_account_impersonation_url: Optional[str] = None

    # token_endpoint
    #   a field that we expect to be a dictionary of values used to create
    #   access tokens from an external identity provider integrated with GCP's
    #   workload identity federation service
    token_endpoint: Optional[Dict[str, str]] = None

    compute_region: Optional[str] = None
    dataproc_cluster_name: Optional[str] = None
    gcs_bucket: Optional[str] = None
    submission_method: Optional[str] = None

    dataproc_batch: Optional[DataprocBatchConfig] = field(
        metadata={
            "serialization_strategy": pass_through,
        },
        default=None,
    )

    scopes: Optional[Tuple[str, ...]] = (
        "https://www.googleapis.com/auth/bigquery",
        "https://www.googleapis.com/auth/cloud-platform",
        "https://www.googleapis.com/auth/drive",
        "https://www.googleapis.com/auth/userinfo.email",
    )

    _ALIASES = {
        # 'legacy_name': 'current_name'
        "project": "database",
        "dataset": "schema",
        "target_project": "target_database",
        "target_dataset": "target_schema",
        "retries": "job_retries",
        "timeout_seconds": "job_execution_timeout_seconds",
        "dataproc_region": "compute_region",
    }

    def __post_init__(self):
        if self.keyfile_json and "private_key" in self.keyfile_json:
            self.keyfile_json["private_key"] = self.keyfile_json["private_key"].replace(
                "\\n", "\n"
            )
        if not self.method:
            raise DbtRuntimeError("Must specify authentication method")

        if not self.schema:
            raise DbtRuntimeError("Must specify schema")

    @property
    def type(self):
        return "bigquery"

    @property
    def unique_field(self):
        return self.database

    def _connection_keys(self):
        return (
            "method",
            "database",
            "execution_project",
            "schema",
            "location",
            "priority",
            "maximum_bytes_billed",
            "impersonate_service_account",
            "job_retry_deadline_seconds",
            "job_retries",
            "job_creation_timeout_seconds",
            "job_execution_timeout_seconds",
            "timeout_seconds",
            "client_id",
            "token_uri",
            "compute_region",
            "dataproc_cluster_name",
            "gcs_bucket",
            "dataproc_batch",
        )

    @classmethod
    def __pre_deserialize__(cls, d: Dict[Any, Any]) -> Dict[Any, Any]:
        # We need to inject the correct value of the database (aka project) at
        # this stage, ref
        # https://github.com/dbt-labs/dbt/pull/2908#discussion_r532927436.

        # `database` is an alias of `project` in BigQuery
        if "database" not in d:
            _, database = _create_bigquery_defaults()
            d["database"] = database
        # `execution_project` default to dataset/project
        if "execution_project" not in d:
            d["execution_project"] = d["database"]
        return d


def set_default_credentials() -> None:
    try:
        run_cmd(".", ["gcloud", "--version"])
    except OSError as e:
        _logger.debug(e)
        msg = """
        dbt requires the gcloud SDK to be installed to authenticate with BigQuery.
        Please download and install the SDK, or use a Service Account instead.

        https://cloud.google.com/sdk/
        """
        raise DbtRuntimeError(msg)

    run_cmd(".", ["gcloud", "auth", "application-default", "login"])


def create_google_credentials(credentials: BigQueryCredentials) -> GoogleCredentials:
    if credentials.impersonate_service_account:
        return _create_impersonated_credentials(credentials)
    return _create_google_credentials(credentials)


def _create_impersonated_credentials(credentials: BigQueryCredentials) -> ImpersonatedCredentials:
    if credentials.scopes and isinstance(credentials.scopes, Iterable):
        target_scopes = list(credentials.scopes)
    else:
        target_scopes = []

    return ImpersonatedCredentials(
        source_credentials=_create_google_credentials(credentials),
        target_principal=credentials.impersonate_service_account,
        target_scopes=target_scopes,
    )


def _create_google_credentials(credentials: BigQueryCredentials) -> GoogleCredentials:

    if credentials.method == BigQueryConnectionMethod.OAUTH:
        creds, _ = _create_bigquery_defaults(scopes=credentials.scopes)

    elif credentials.method == BigQueryConnectionMethod.SERVICE_ACCOUNT:
        creds = ServiceAccountCredentials.from_service_account_file(
            credentials.keyfile, scopes=credentials.scopes
        )

    elif credentials.method == BigQueryConnectionMethod.SERVICE_ACCOUNT_JSON:
        details = credentials.keyfile_json
        if _is_base64(details):  # type:ignore
            details = _base64_to_string(details)
        creds = ServiceAccountCredentials.from_service_account_info(
            details, scopes=credentials.scopes
        )

    elif credentials.method == BigQueryConnectionMethod.OAUTH_SECRETS:
        creds = GoogleCredentials(
            token=credentials.token,
            refresh_token=credentials.refresh_token,
            client_id=credentials.client_id,
            client_secret=credentials.client_secret,
            token_uri=credentials.token_uri,
            scopes=credentials.scopes,
        )

    elif credentials.method == BigQueryConnectionMethod.EXTERNAL_OAUTH_WIF:
        creds = _create_identity_pool_credentials(credentials=credentials)

    else:
        raise FailedToConnectError(f"Invalid `method` in profile: '{credentials.method}'")

    return creds


def _create_identity_pool_credentials(credentials: BigQueryCredentials) -> GoogleCredentials:
    if not credentials.token_endpoint:
        raise FailedToConnectError("token_endpoint is required for external-oauth-wif")
    token_supplier = create_token_supplier(credentials.token_endpoint)

    # The dict here represents an Application Default Credentials configuration.
    # Identity pool credentials can be created from these configurations, which informs the Google clients:
    #   1. How to retrieve an access token from an external IdP, which is specified in the `credential_source` blob
    #   2. Where to exchange that access token for a short-lived security token, which is specified by the
    #      `token_url`; this should point to Google's Security Token Service (STS) API
    #   3. The intended audience of the short-lived security tokens issued by Google's STS API,
    #      which is generally the fully specified resource name of the workload pool provider
    adc_dict = {
        "universe_domain": "googleapis.com",
        "type": "external_account",
        "audience": credentials.workload_pool_provider_path,
        "subject_token_type": "urn:ietf:params:oauth:token-type:jwt",
        "token_url": "https://sts.googleapis.com/v1/token",
        "subject_token_supplier": token_supplier,
    }

    # The service account impersonation URL here is optional, but we expect to see it in cases where IAM roles have not been
    # assigned to external identities (such as Entra) or direct resource access has not been granted to the workpool
    if credentials.service_account_impersonation_url:
        adc_dict["service_account_impersonation_url"] = (
            credentials.service_account_impersonation_url
        )

    creds = IdentityPoolCredentials.from_info(adc_dict)
    return creds.with_scopes(credentials.scopes)


@lru_cache()
def _create_bigquery_defaults(scopes=None) -> Tuple[Any, Optional[str]]:
    """
    Returns (credentials, project_id)

    project_id is returned available from the environment; otherwise None
    """
    # Cached, because the underlying implementation shells out, taking ~1s
    try:
        return default(scopes=scopes)
    except DefaultCredentialsError as e:
        raise DbtConfigError(f"Failed to authenticate with supplied credentials\nerror:\n{e}")


def _is_base64(s: Union[str, bytes]) -> bool:
    """
    Checks if the given string or bytes object is valid Base64 encoded.

    Args:
        s: The string or bytes object to check.

    Returns:
        True if the input is valid Base64, False otherwise.
    """

    if isinstance(s, str):
        # For strings, ensure they consist only of valid Base64 characters
        if not s.isascii():
            return False
        # Convert to bytes for decoding
        s = s.encode("ascii")

    try:
        # Use the 'validate' parameter to enforce strict Base64 decoding rules
        base64.b64decode(s, validate=True)
        return True
    except (TypeError, binascii.Error):
        return False


def _base64_to_string(b):
    return base64.b64decode(b).decode("utf-8")



import pytest

from google.cloud.bigquery.client import Client

from dbt.tests.util import run_dbt


_MACRO__BQ_LABELS = """
{% macro bq_labels() %}{
    "system": "{{ env_var('LABEL_SYSTEM', 'my_system') }}",
    "env_type": "{{ env_var('LABEL_ENV', 'dev') }}"
}{% endmacro %}
"""
_MODEL__MY_TABLE = """
{{ config(materialized= "table") }}
select 1 as id
"""


class TestQueryCommentJobLabel:
    @pytest.fixture(scope="class")
    def models(self):
        return {"my_table.sql": _MODEL__MY_TABLE}

    @pytest.fixture(scope="class")
    def macros(self):
        return {"bq_labels.sql": _MACRO__BQ_LABELS}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "query-comment": {
                "comment": "{{ bq_labels() }}",
                "job-label": True,
                "append": True,
            }
        }

    def test_query_comments_displays_as_job_labels(self, project):
        """
        Addresses this regression in dbt-bigquery 1.6:
        https://github.com/dbt-labs/dbt-bigquery/issues/863
        """
        results = run_dbt(["run"])
        job_id = results.results[0].adapter_response.get("job_id")
        with project.adapter.connection_named("_test"):
            client: Client = project.adapter.connections.get_thread_connection().handle
            job = client.get_job(job_id=job_id)

        # this is what should happen
        assert job.labels.get("system") == "my_system"
        assert job.labels.get("env_type") == "dev"



import copy
from typing import Any, Dict, List, Mapping, Sequence

from dbt.adapters.bigquery.column import BigQueryColumn


def merge_nested_fields(
    existing_fields: Sequence[Dict[str, Any]],
    additions: Mapping[str, Dict[str, Any]],
    prefix: str = "",
) -> List[Dict[str, Any]]:
    """Merge new fields into existing STRUCT fields, appending at each nesting level.

    Note: Primarily used for field removal. For adding fields, sync_struct_columns
    uses the source schema directly to preserve field order.
    """
    merged_fields: List[Dict[str, Any]] = []
    addition_lookup = dict(additions)

    for field in existing_fields:
        field_name = field["name"]
        qualified_name = f"{prefix}.{field_name}" if prefix else field_name

        direct_addition = addition_lookup.pop(qualified_name, None)
        if direct_addition is not None:
            merged_fields.append(copy.deepcopy(direct_addition))
            continue

        nested_additions = {
            key: value
            for key, value in list(addition_lookup.items())
            if key.startswith(f"{qualified_name}.")
        }

        if nested_additions and field.get("type") == "RECORD":
            for key in nested_additions:
                addition_lookup.pop(key, None)

            stripped_additions = {
                key.split(".", 1)[1]: value for key, value in nested_additions.items()
            }

            merged_children = merge_nested_fields(
                field.get("fields", []) or [],
                stripped_additions,
                prefix="",
            )

            merged_field = copy.deepcopy(field)
            merged_field["fields"] = merged_children
            merged_fields.append(merged_field)
        else:
            merged_fields.append(copy.deepcopy(field))

    for path, addition in addition_lookup.items():
        if "." not in path:
            merged_fields.append(copy.deepcopy(addition))

    return merged_fields


def collect_field_dicts(
    fields: Sequence[Dict[str, Any]], prefix: str = ""
) -> Dict[str, Dict[str, Any]]:
    collected: Dict[str, Dict[str, Any]] = {}
    for field in fields:
        name = field["name"]
        path = f"{prefix}.{name}" if prefix else name
        collected[path] = field
        if field.get("type") == "RECORD":
            collected.update(collect_field_dicts(field.get("fields", []) or [], path))
    return collected


def find_missing_fields(
    source_fields: Sequence[Dict[str, Any]],
    target_fields: Sequence[Dict[str, Any]],
) -> Dict[str, Dict[str, Any]]:
    source_map = collect_field_dicts(source_fields)
    target_map = collect_field_dicts(target_fields)
    return {
        path: copy.deepcopy(field) for path, field in source_map.items() if path not in target_map
    }


def build_nested_additions(add_columns: Sequence[BigQueryColumn]) -> Dict[str, Dict[str, Any]]:
    additions: Dict[str, Dict[str, Any]] = {}

    for column in add_columns:
        schema_field = column.column_to_bq_schema().to_api_repr()
        additions[column.name] = schema_field

    return additions



from datetime import datetime
import os
import random

from dbt.tests.util import get_connection, run_dbt
import pytest

from dbt.adapters.bigquery import BigQueryRelation


def prefix():
    # create a directory name that will be unique per test session
    _randint = random.randint(0, 9999)
    _runtime_timedelta = datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0)
    _runtime = (int(_runtime_timedelta.total_seconds() * 1e6)) + _runtime_timedelta.microseconds
    return f"test{_runtime}{_randint:04}"


# make sure this is static
PREFIX = prefix()


MODEL__INFO_SCHEMA_DEFAULT = """
select 1 as id
"""
MODEL__INFO_SCHEMA_TABLE = """
{{ config(materialized='table') }}
select 1 as id
"""
MODEL__MANAGED_ICEBERG_TABLE = (
    """
{{ config(
    materialized='table',
    catalog_name='managed_iceberg',
    storage_uri='gs://"""
    + os.getenv("BIGQUERY_TEST_ICEBERG_BUCKET")
    + """/"""
    + PREFIX
    + """__managed_iceberg_table'
) }}
select 1 as id
"""
)
MODEL__MANAGED_ICEBERG_INCREMENTAL = (
    """
{{ config(
    materialized="incremental",
    unique_key="id",
    catalog_name='managed_iceberg',
    storage_uri='gs://"""
    + os.getenv("BIGQUERY_TEST_ICEBERG_BUCKET")
    + """/"""
    + PREFIX
    + """__managed_iceberg_incremental'
) }}

with data as (

    {% if not is_incremental() %}

        select 1 as id, cast('2020-01-01' as datetime) as date_time union all
        select 2 as id, cast('2020-01-01' as datetime) as date_time union all
        select 3 as id, cast('2020-01-01' as datetime) as date_time union all
        select 4 as id, cast('2020-01-01' as datetime) as date_time

    {% else %}

        select 1 as id, cast('2020-01-01' as datetime) as date_time union all
        select 2 as id, cast('2020-01-01' as datetime) as date_time union all
        select 3 as id, cast('2020-01-01' as datetime) as date_time union all
        select 4 as id, cast('2020-01-02' as datetime) as date_time union all
        select 5 as id, cast('2020-01-02' as datetime) as date_time union all
        select 6 as id, cast('2020-01-02' as datetime) as date_time

    {% endif %}

)

select * from data

{% if is_incremental() %}
where id >= (select max(id) from {{ this }})
{% endif %}
"""
)


class TestGenericCatalog:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "info_schema_default.sql": MODEL__INFO_SCHEMA_DEFAULT,
            "info_schema_table.sql": MODEL__INFO_SCHEMA_TABLE,
            "managed_iceberg_table.sql": MODEL__MANAGED_ICEBERG_TABLE,
            "managed_iceberg_incremental.sql": MODEL__MANAGED_ICEBERG_INCREMENTAL,
        }

    def test_generic_catalog(self, project):
        results = run_dbt(["run"])
        assert len(results) == 4
        results = run_dbt(["run", "--select", "managed_iceberg_incremental"])
        assert len(results) == 1

        schema = BigQueryRelation.create(project.database, project.test_schema)
        with get_connection(project.adapter):
            relations = project.adapter.list_relations_without_caching(schema)
        assert len(relations) == 4



import pytest

from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.contracts.relation import RelationType
from dbt.tests.util import get_connection, run_dbt

from dbt.adapters.bigquery.relation_configs import BigQueryMaterializedViewConfig
from tests.functional.adapter.describe_relation import _files


class TestDescribeRelation:
    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": _files.MY_SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            "my_base_table.sql": _files.MY_BASE_TABLE,
            "my_materialized_view.sql": _files.MY_MATERIALIZED_VIEW,
            "my_other_base_table.sql": _files.MY_OTHER_BASE_TABLE,
            "my_other_materialized_view.sql": _files.MY_OTHER_MATERIALIZED_VIEW,
        }

    @pytest.fixture(scope="class")
    def my_seed(self, project) -> BaseRelation:
        return project.adapter.Relation.create(
            identifier="my_seed",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.Table,
        )

    @pytest.fixture(scope="class")
    def my_base_table(self, project) -> BaseRelation:
        """
        The base table for a materialized view needs to be partitioned in
        the same way as the materialized view. So if we want to create a partitioned
        materialized view, we need to partition the base table. This table is a
        select * on the seed table, plus a partition.
        """
        return project.adapter.Relation.create(
            identifier="my_base_table",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.Table,
        )

    @pytest.fixture(scope="class")
    def my_materialized_view(self, project) -> BaseRelation:
        return project.adapter.Relation.create(
            identifier="my_materialized_view",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.MaterializedView,
        )

    @pytest.fixture(scope="class")
    def my_other_materialized_view(self, project) -> BaseRelation:
        return project.adapter.Relation.create(
            identifier="my_other_materialized_view",
            schema=project.test_schema,
            database=project.database,
            type=RelationType.MaterializedView,
        )

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project, my_base_table, my_materialized_view):
        run_dbt(["seed"])
        run_dbt(["run"])
        yield
        project.run_sql(f"drop schema if exists {project.test_schema} cascade")

    def test_describe_materialized_view(self, project, my_materialized_view):
        with get_connection(project.adapter):
            results = project.adapter.describe_relation(my_materialized_view)
        assert isinstance(results, BigQueryMaterializedViewConfig)
        assert results.table_id == f'"{my_materialized_view.identifier}"'
        assert results.dataset_id == f'"{my_materialized_view.schema}"'
        assert results.project_id == f'"{my_materialized_view.database}"'
        assert results.partition.field == "record_date"
        assert results.partition.data_type == "datetime"
        assert results.partition.granularity == "day"
        assert results.cluster.fields == frozenset({"id"})
        assert results.options.enable_refresh is True
        assert results.options.refresh_interval_minutes == 30

    def test_describe_other_materialized_view(self, project, my_other_materialized_view):
        with get_connection(project.adapter):
            results = project.adapter.describe_relation(my_other_materialized_view)
        assert isinstance(results, BigQueryMaterializedViewConfig)
        assert results.table_id == f'"{my_other_materialized_view.identifier}"'
        assert results.dataset_id == f'"{my_other_materialized_view.schema}"'
        assert results.project_id == f'"{my_other_materialized_view.database}"'
        assert results.partition.field == "value"
        assert results.partition.data_type == "int64"
        assert results.partition.range == {"start": 0, "end": 500, "interval": 50}
        assert results.cluster.fields == frozenset({"id"})
        assert results.options.enable_refresh is False
        assert results.options.refresh_interval_minutes == 30  # BQ returns it to the default
